# mypy: ignore-errors
from pathlib import Path
from typing import Any, List

from fastmcp import FastMCP
from mcp.types import ToolAnnotations
from pydantic.dataclasses import dataclass

from memalot.reports import (
    FullReport,
    ReportSummary,
    filter_iteration_by_types,
    get_report_reader,
)

mcp: FastMCP[Any] = FastMCP("Memalot MCP Server")


@dataclass
class ReportSummaries:
    """
    Wraps a list of report summaries so that they can be returning from the MCP server.
    Just returning a list gives slightly odd results.
    """

    summaries: List[ReportSummary]


@mcp.tool(
    annotations=ToolAnnotations(
        title="List Reports",
        readOnlyHint=True,
        destructiveHint=False,
        idempotentHint=True,
        openWorldHint=False,
    )
)
def list_leak_reports(num_reports: int = 5, report_directory: str | None = None) -> ReportSummaries:
    """
    List the `num_reports` most recent memory reports that have been run.
    The reports are ordered by start time with the most recent report first.

    By default, the server will return the 5 most recent reports.
    This can be overridden by passing a different value for `num_reports`.

    Args:
        num_reports: The number of reports to return.
        report_directory: The directory to search for reports. If `None`, the
        default directory will be used. This is the `.memalot/reports` directory
        in the user's home directory.

    Returns:
        A `ReportSummaries` object.
    """
    reader = get_report_reader(
        report_directory=Path(report_directory) if report_directory else None
    )
    summaries = reader.get_report_summaries()
    summaries.sort(key=lambda s: s.metadata.start_time, reverse=True)
    return ReportSummaries(summaries[:num_reports])


@mcp.tool(
    annotations=ToolAnnotations(
        title="Get Report",
        readOnlyHint=True,
        destructiveHint=False,
        idempotentHint=True,
        openWorldHint=False,
    )
)
def get_leak_report(
    report_id: str,
    num_iterations: int = 1,
    report_directory: str | None = None,
    filter_types: str | None = None,
) -> FullReport:
    """
    Get the full report for a specific ID.

    By default, the server will return the most recent iteration of the report.
    This can be overridden by passing a different value for `num_iterations`.

    Args:
        report_id: The ID of the report to get. This can be obtained from the
        `list_reports` method.
        num_iterations: The number of iterations to return.
        report_directory: The directory to search for reports. If `None`, the
        default directory will be used.
        filter_types: A comma-separated list of types to filter on. Only these types
        are included in the report. This is useful for reducing the size of reports
        to more easily fit them in a model's context window, for example.

    Returns:
        A `FullReport` object.
    """
    reader = get_report_reader(
        report_directory=Path(report_directory) if report_directory else None
    )
    full_report = reader.get_full_report(report_id=report_id, num_iterations=num_iterations)

    if filter_types:
        filter_types_list = [t.strip() for t in filter_types.split(",") if t.strip()]
        # Filter each iteration
        filtered_iterations = [
            filter_iteration_by_types(iteration, filter_types_list)
            for iteration in full_report.iterations
        ]
        # Return a new FullReport with filtered iterations
        return FullReport(
            summary=full_report.summary,
            iterations=filtered_iterations,
        )
    else:
        return full_report


def run_mcp_server() -> None:  # pragma: no cover
    """
    Run the Memalot MCP server.
    """
    mcp.run()


if __name__ == "__main__":  # pragma: no cover
    run_mcp_server()



from datetime import datetime, timezone
from typing import Any
from unittest.mock import MagicMock

import pytest

from memalot.base import ApproximateSize, CachingIterable, MemalotCount
from memalot.memory import MemalotMemoryUsage
from memalot.options import Options
from memalot.output import Output, OutputWriter
from memalot.report_generator import ReportGenerator
from memalot.reports import (
    LeakSummary,
    MemoryUsageOutput,
    ObjectDetails,
    ReportIteration,
    ReportWriter,
    TypeSummary,
)
from memalot.snapshots import MemalotObjects, MemalotSnapshotManager
from tests.utils_for_testing import create_mock


class _FakeOutputWriter(OutputWriter):
    """
    Simple OutputWriter implementation for testing that captures written output.
    """

    def __init__(self) -> None:
        self.written_outputs: list[ReportIteration] = []

    def write(self, output: Output) -> None:
        # For testing, we expect ReportIteration which is a subclass of Output
        assert isinstance(output, ReportIteration)
        self.written_outputs.append(output)


class _FakeReportWriter(ReportWriter):
    """
    Simple ReportWriter implementation for testing that captures written reports.
    """

    def __init__(self, report_id: str) -> None:
        self._report_id = report_id
        self.written_reports: list[ReportIteration] = []

    def write_iteration(self, iteration: ReportIteration) -> None:
        self.written_reports.append(iteration)

    @property
    def report_id(self) -> str:
        return self._report_id


@pytest.fixture(name="report_generator")
def _report_generator() -> ReportGenerator:
    """
    Creates a ReportGenerator instance for testing.
    """
    return ReportGenerator()


@pytest.fixture(name="mock_iteration")
def _mock_iteration() -> MemalotCount:
    """
    Creates a MemalotCount instance for testing.
    """
    return MemalotCount(3)


@pytest.fixture(name="mock_type_summaries")
def _mock_type_summaries() -> list[TypeSummary]:
    return [
        TypeSummary(
            object_type="builtins.str",
            count=5,
            shallow_size_bytes=ApproximateSize(approx_size=100, upper_bound_known=True),
        ),
        TypeSummary(
            object_type="builtins.int",
            count=3,
            shallow_size_bytes=ApproximateSize(approx_size=72, upper_bound_known=True),
        ),
    ]


@pytest.fixture(name="mock_leak_summary")
def _mock_leak_summary(
    mock_iteration: MemalotCount, mock_type_summaries: list[TypeSummary]
) -> LeakSummary:
    return LeakSummary(
        iteration=mock_iteration,
        type_summaries=mock_type_summaries,
    )


@pytest.fixture(name="mock_object_details")
def _mock_object_details() -> list[ObjectDetails]:
    return [create_mock(spec=ObjectDetails), create_mock(spec=ObjectDetails)]


@pytest.fixture(name="mock_memory_usage")
def _mock_memory_usage(mock_iteration: MemalotCount) -> MemalotMemoryUsage:
    return MemalotMemoryUsage(
        current_rss_bytes=1024 * 1024,
        peak_rss_bytes=2048 * 1024,
        system_percent_used=25.5,
        iteration_number=int(mock_iteration),
    )


@pytest.fixture(name="mock_objects_with_details")
def _mock_objects_with_details(
    mock_leak_summary: LeakSummary,
    mock_object_details: list[ObjectDetails],
) -> MemalotObjects:
    """
    Creates MemalotObjects with mocked details.
    """
    mock_objects = create_mock(spec=MemalotObjects)
    mock_objects.generate_object_details.return_value = iter(mock_object_details)
    mock_objects.get_leak_summary.return_value = mock_leak_summary
    mock_objects.__len__.return_value = 8
    return mock_objects


@pytest.fixture(name="mock_empty_objects")
def _mock_empty_objects() -> MemalotObjects:
    """
    Creates empty MemalotObjects.
    """
    mock_objects = create_mock(spec=MemalotObjects)
    mock_objects.generate_object_details.return_value = iter([])
    mock_objects.get_leak_summary.return_value = create_mock(spec=LeakSummary)
    mock_objects.__len__.return_value = 0
    return mock_objects


@pytest.fixture(name="mock_memory_usage_provider")
def _mock_memory_usage_provider(
    mock_memory_usage: MemalotMemoryUsage,
) -> MagicMock:
    """
    Creates a mock memory usage provider for testing.
    """
    mock_provider = create_mock(spec=MemalotSnapshotManager)
    mock_provider.rotate_memory_usage.return_value = None, mock_memory_usage
    return mock_provider


@pytest.fixture(name="mock_memory_usage_provider_with_previous")
def _mock_memory_usage_provider_with_previous(
    mock_memory_usage: MemalotMemoryUsage,
    mock_iteration: MemalotCount,
) -> MagicMock:
    """
    Creates a mock memory usage provider with previous memory usage.
    """
    mock_provider = create_mock(spec=MemalotSnapshotManager)
    mock_provider.rotate_memory_usage.return_value = (
        MemalotMemoryUsage(
            current_rss_bytes=0,
            peak_rss_bytes=0,
            system_percent_used=0,
            iteration_number=int(mock_iteration) - 1,
        ),
        mock_memory_usage,
    )
    return mock_provider


@pytest.fixture(name="fake_output_writer")
def _fake_output_writer() -> _FakeOutputWriter:
    """
    Creates a fake output writer that captures written output.
    """
    return _FakeOutputWriter()


@pytest.fixture(name="fake_report_writer")
def _fake_report_writer() -> _FakeReportWriter:
    """
    Creates a fake report writer that captures written reports.
    """
    return _FakeReportWriter(report_id="test-report-123")


@pytest.fixture(name="default_options")
def _default_options() -> Options:
    """
    Creates default options for testing.
    """
    return Options()


@pytest.fixture(name="test_objects")
def _test_objects() -> list[Any]:
    """
    Creates a list of test objects.
    """
    return ["test_string", 42, [1, 2, 3], {"key": "value"}]


class TestReportGenerator:
    """
    Tests for the `ReportGenerator` class.
    """

    def test_generate_report(
        self,
        report_generator: ReportGenerator,
        mock_memory_usage_provider: MagicMock,
        fake_output_writer: _FakeOutputWriter,
        fake_report_writer: _FakeReportWriter,
        default_options: Options,
        test_objects: list[Any],
        mock_iteration: MemalotCount,
        mock_memory_usage: MemalotMemoryUsage,
    ) -> None:
        """
        Tests the basic flow of the generate_report method with detailed_report=True.
        """
        excluded_from_referrers: list[int] = []

        before_time = datetime.now(timezone.utc)
        iteration_start_time = datetime.now(timezone.utc)
        report_generator.generate_report(
            report_id="test-report-123",
            iteration_start_time=iteration_start_time,
            memory_usage_provider=mock_memory_usage_provider,
            objects=MemalotObjects(test_objects),
            output_writer=fake_output_writer,
            report_writer=fake_report_writer,
            options=default_options,
            iteration=mock_iteration,
            excluded_from_referrers=excluded_from_referrers,
            detailed_report=True,
        )
        after_time = datetime.now(timezone.utc)

        assert len(fake_output_writer.written_outputs) == 1
        assert len(fake_report_writer.written_reports) == 1

        written_output = fake_output_writer.written_outputs[0]
        written_report = fake_report_writer.written_reports[0]

        # Both writers should have the same object (at the moment)
        assert written_output is written_report

        assert written_report.report_id == "test-report-123"
        assert written_report.iteration_number == mock_iteration
        assert before_time <= written_report.end_time <= after_time

        expected_memory_usage = MemoryUsageOutput(
            current_rss_bytes=mock_memory_usage.current_rss_bytes,
            peak_rss_bytes=mock_memory_usage.peak_rss_bytes,
            system_percent_used=mock_memory_usage.system_percent_used,
            current_rss_diff=None,  # No previous usage
            peak_rss_diff=None,
            diff_from_iteration=None,
        )
        assert written_report.memory_usage == expected_memory_usage

        # Check the leak summary has correct types (based on test_objects)
        assert written_report.leak_summary.iteration == mock_iteration
        type_names = [ts.object_type for ts in written_report.leak_summary.type_summaries]
        # test_objects = ["test_string", 42, [1, 2, 3], {"key": "value"}]
        assert "builtins.str" in type_names
        assert "builtins.int" in type_names
        assert "builtins.list" in type_names
        assert "builtins.dict" in type_names

        assert isinstance(written_report.object_details_list, CachingIterable)

        # Verify memory usage was rotated
        mock_memory_usage_provider.rotate_memory_usage.assert_called_once_with(
            iteration=mock_iteration
        )

    def test_generate_report_no_snapshot(
        self,
        report_generator: ReportGenerator,
        mock_memory_usage_provider: MagicMock,
        fake_output_writer: _FakeOutputWriter,
        fake_report_writer: _FakeReportWriter,
        default_options: Options,
        test_objects: list[Any],
        mock_iteration: MemalotCount,
        mock_memory_usage: MemalotMemoryUsage,
    ) -> None:
        """
        Tests the generate_report method and detailed_report=True.

        Currently, this just generates an empty report (no error).
        """
        excluded_from_referrers: list[int] = []
        iteration_start_time = datetime.now(timezone.utc)

        report_generator.generate_report(
            report_id="test-report-123",
            iteration_start_time=iteration_start_time,
            memory_usage_provider=mock_memory_usage_provider,
            objects=MemalotObjects(test_objects),
            output_writer=fake_output_writer,
            report_writer=fake_report_writer,
            options=default_options,
            iteration=mock_iteration,
            excluded_from_referrers=excluded_from_referrers,
            detailed_report=True,
        )

        # Should generate a report
        written_report = fake_report_writer.written_reports[0]
        assert written_report.report_id == "test-report-123"

    def test_generate_report_no_diff(
        self,
        report_generator: ReportGenerator,
        mock_memory_usage_provider: MagicMock,
        fake_output_writer: _FakeOutputWriter,
        fake_report_writer: _FakeReportWriter,
        default_options: Options,
        test_objects: list[Any],
        mock_iteration: MemalotCount,
        mock_memory_usage: MemalotMemoryUsage,
    ) -> None:
        """
        Tests the generate_report method with empty objects and detailed_report=True.
        """
        excluded_from_referrers: list[int] = []
        iteration_start_time = datetime.now(timezone.utc)

        report_generator.generate_report(
            report_id="test-report-123",
            iteration_start_time=iteration_start_time,
            memory_usage_provider=mock_memory_usage_provider,
            objects=MemalotObjects([]),
            output_writer=fake_output_writer,
            report_writer=fake_report_writer,
            options=default_options,
            iteration=mock_iteration,
            excluded_from_referrers=excluded_from_referrers,
            detailed_report=True,
        )
        written_report = fake_report_writer.written_reports[0]

        assert written_report.leak_summary == LeakSummary(
            iteration=mock_iteration,
            type_summaries=[],
            max_types_in_summary=500,
        )

        assert isinstance(written_report.object_details_list, CachingIterable)
        assert list(written_report.object_details_list) == []

        # Verify memory usage was rotated
        mock_memory_usage_provider.rotate_memory_usage.assert_called_once_with(
            iteration=mock_iteration
        )

    def test_generate_report_with_previous_memory_usage(
        self,
        report_generator: ReportGenerator,
        mock_memory_usage_provider_with_previous: MagicMock,
        fake_output_writer: _FakeOutputWriter,
        fake_report_writer: _FakeReportWriter,
        default_options: Options,
        test_objects: list[Any],
        mock_iteration: MemalotCount,
        mock_memory_usage: MemalotMemoryUsage,
    ) -> None:
        """
        Tests the generate_report method with previous memory usage present and
        detailed_report=True.
        """
        excluded_from_referrers: list[int] = []
        iteration_start_time = datetime.now(timezone.utc)

        report_generator.generate_report(
            report_id="test-report-123",
            iteration_start_time=iteration_start_time,
            memory_usage_provider=mock_memory_usage_provider_with_previous,
            objects=MemalotObjects([]),
            output_writer=fake_output_writer,
            report_writer=fake_report_writer,
            options=default_options,
            iteration=mock_iteration,
            excluded_from_referrers=excluded_from_referrers,
            detailed_report=True,
        )

        written_report = fake_report_writer.written_reports[0]

        assert written_report.memory_usage == MemoryUsageOutput(
            current_rss_bytes=mock_memory_usage.current_rss_bytes,
            peak_rss_bytes=mock_memory_usage.peak_rss_bytes,
            system_percent_used=mock_memory_usage.system_percent_used,
            current_rss_diff=mock_memory_usage.current_rss_bytes,
            peak_rss_diff=mock_memory_usage.peak_rss_bytes,
            diff_from_iteration=mock_iteration - 1,
        )

    def test_generate_report_detailed_report_false(
        self,
        report_generator: ReportGenerator,
        mock_memory_usage_provider: MagicMock,
        fake_output_writer: _FakeOutputWriter,
        fake_report_writer: _FakeReportWriter,
        default_options: Options,
        test_objects: list[Any],
        mock_iteration: MemalotCount,
        mock_memory_usage: MemalotMemoryUsage,
    ) -> None:
        """
        Tests the generate_report method with detailed_report=False returns empty summary
        and referrers.
        """
        excluded_from_referrers: list[int] = []

        before_time = datetime.now(timezone.utc)
        iteration_start_time = datetime.now(timezone.utc)
        report_generator.generate_report(
            report_id="test-report-123",
            iteration_start_time=iteration_start_time,
            memory_usage_provider=mock_memory_usage_provider,
            objects=MemalotObjects(test_objects),
            output_writer=fake_output_writer,
            report_writer=fake_report_writer,
            options=default_options,
            iteration=mock_iteration,
            excluded_from_referrers=excluded_from_referrers,
            detailed_report=False,
        )
        after_time = datetime.now(timezone.utc)

        assert len(fake_output_writer.written_outputs) == 1
        assert len(fake_report_writer.written_reports) == 1

        written_output = fake_output_writer.written_outputs[0]
        written_report = fake_report_writer.written_reports[0]

        # Both writers should have the same object (at the moment)
        assert written_output is written_report

        assert written_report.report_id == "test-report-123"
        assert written_report.iteration_number == mock_iteration
        assert before_time <= written_report.end_time <= after_time

        expected_memory_usage = MemoryUsageOutput(
            current_rss_bytes=mock_memory_usage.current_rss_bytes,
            peak_rss_bytes=mock_memory_usage.peak_rss_bytes,
            system_percent_used=mock_memory_usage.system_percent_used,
            current_rss_diff=None,  # No previous usage
            peak_rss_diff=None,
            diff_from_iteration=None,
        )
        assert written_report.memory_usage == expected_memory_usage

        # When detailed_report=False, should get empty leak summary
        expected_leak_summary = LeakSummary(
            iteration=mock_iteration,
            type_summaries=[],
            max_types_in_summary=default_options.max_types_in_leak_summary,
        )
        assert written_report.leak_summary == expected_leak_summary

        # When detailed_report=False, should get empty referrers
        assert isinstance(written_report.object_details_list, CachingIterable)
        assert list(written_report.object_details_list) == []

        # Verify memory usage was rotated
        mock_memory_usage_provider.rotate_memory_usage.assert_called_once_with(
            iteration=mock_iteration
        )



from typing import List

from rich.console import Console, RenderableType
from rich.console import NewLine as RichNewLine
from rich.text import Text

from memalot.options import Options
from memalot.output import (
    NewBlock,
    NewLine,
    Output,
    RichConsoleWriter,
    RichOutput,
    get_output_writer,
)
from tests.utils_for_testing import create_mock


class TestsRichOutput:
    """
    Tests for the RichOutput class.
    """

    def test_rich_output_init(self) -> None:
        """
        Test that the RichOutput class is initialized correctly.
        """
        output = RichOutput(renderable=Text("Hello, world!"))
        assert output.to_renderables() == [Text("Hello, world!")]


class TestNewLine:
    """
    Tests for the NewLine class.
    """

    def test_new_line(self) -> None:
        """
        Test that the NewLine class functions as expected.
        """
        new_line = NewLine()
        renderables = list(new_line.to_renderables())
        assert len(renderables) == 1
        assert isinstance(renderables[0], RichNewLine)


class TestRichConsoleWriter:
    def test_write__no_output_func(self) -> None:
        """
        Test that the RichConsoleWriter class writes to the console when no output
        function is provided.
        """
        mock_console = create_mock(spec=Console)
        writer = RichConsoleWriter(options=Options(), console=mock_console)
        writer.write(RichOutput(renderable=Text("Hello, world!")))
        mock_console.print.assert_called_once_with(
            Text("Hello, world!"), soft_wrap=True, crop=False, overflow="ignore"
        )

    def test_write__with_output_func(self) -> None:
        """
        Test that the RichConsoleWriter class writes to the console when an output
        function is provided.
        """
        outputs: List[str] = []
        options = Options(output_func=outputs.append)
        writer = get_output_writer(options=options)
        writer.write(RichOutput(renderable=Text("Hello, world!")))
        assert outputs == ["\nHello, world!\n"]

    def test_write__with_output_func__newlines_ignored(self) -> None:
        """
        Test that the RichConsoleWriter class ignores newline renderables when an
        output function is provided.
        """
        outputs: List[str] = []
        options = Options(output_func=outputs.append)
        writer = get_output_writer(options=options)

        class MyOutput(Output):
            def to_renderables(self) -> List[RenderableType]:
                return [Text("Hello, world!"), RichNewLine(), "\n", RichNewLine(), "\n"]

        writer.write(MyOutput())
        assert outputs == ["\nHello, world!\n"]

    def test_write__with_new_block(self) -> None:
        """
        Test that the RichConsoleWriter handles NewBlock renderables correctly.
        """
        outputs: List[str] = []
        options = Options(output_func=outputs.append)
        writer = get_output_writer(options=options)

        class MyOutput(Output):
            def to_renderables(self) -> List[RenderableType]:
                return [Text("First block"), NewBlock("separator"), Text("Second block")]

        writer.write(MyOutput())
        # Should have two separate outputs due to NewBlock separator
        assert len(outputs) == 2
        assert "First block" in outputs[0]
        assert "Second block" in outputs[1]



from pathlib import Path

import pytest
from _pytest.capture import CaptureFixture

from memalot.cli import entrypoint


class TestListReports:
    """
    Functional tests for the "list" cli command.
    """

    @pytest.fixture(name="test_reports_dir")
    def _test_reports_dir(self) -> Path:
        """
        Fixture providing the path to test reports directory.
        """
        return Path(__file__).parent / "test_data" / "reports" / "v1"

    def test_list_command_basic(self, capsys: CaptureFixture[str], test_reports_dir: Path) -> None:
        """
        Test basic list command functionality with existing test data.
        """
        # Act
        result = entrypoint(["list", "--report-directory", str(test_reports_dir)])

        # Assert
        assert result == 0
        captured = capsys.readouterr()

        assert "Reports" in captured.out
        assert "Report ID" in captured.out
        assert "Iterations" in captured.out
        assert "Start Time" in captured.out
        assert "Entrypoint" in captured.out
        assert "Arguments" in captured.out

        assert "d6wf-uw2w" in captured.out
        assert "2025-09-29 20:54:18" in captured.out
        assert "utils_for_tes" in captured.out  # Truncated in table display

        assert "lyy0-uf3p" in captured.out
        assert "2025-09-29 20:54:06" in captured.out
        assert "utils_for_tes" in captured.out  # Truncated in table display

        assert "ykik-ab4l" in captured.out
        assert "2025-09-29 20:53:39" in captured.out
        assert "utils_for_tes" in captured.out  # Truncated in table display
        assert "dummy_arg1" in captured.out
        assert "dummy_arg2" in captured.out

    def test_list_command_with_num_reports(
        self, capsys: CaptureFixture[str], test_reports_dir: Path
    ) -> None:
        """
        Test list command with --num-reports option.
        """
        # Act
        result = entrypoint(
            ["list", "--num-reports", "2", "--report-directory", str(test_reports_dir)]
        )

        # Assert
        assert result == 0
        captured = capsys.readouterr()

        # Should show the most recent two reports from test data
        assert "d6wf-uw2w" in captured.out
        assert "lyy0-uf3p" in captured.out
        assert "ykik-ab4l" not in captured.out

    def test_list_command_empty_directory(
        self, capsys: CaptureFixture[str], tmp_path: Path
    ) -> None:
        """
        Test list command with empty report directory.
        """
        # Create empty directory
        empty_dir = tmp_path / "empty_reports"
        empty_dir.mkdir()

        # Act
        result = entrypoint(["list", "--report-directory", str(empty_dir)])

        # Assert
        assert result == 0
        captured = capsys.readouterr()

        # Should show "no reports found" message
        assert "No reports found" in captured.out

    def test_list_command_nonexistent_directory(
        self, capsys: CaptureFixture[str], tmp_path: Path
    ) -> None:
        """
        Test list command with nonexistent report directory.
        """
        nonexistent_dir = tmp_path / "does_not_exist"

        # Act
        result = entrypoint(["list", "--report-directory", str(nonexistent_dir)])

        # Assert
        assert result == 0
        captured = capsys.readouterr()

        # Should show "no reports found" message since directory doesn't exist
        assert "No reports found" in captured.out

    def test_list_command_high_num_reports(
        self, capsys: CaptureFixture[str], test_reports_dir: Path
    ) -> None:
        """
        Test list command with higher --num-reports than available reports.
        """
        # Act - request more reports than exist in test data
        result = entrypoint(
            ["list", "--num-reports", "100", "--report-directory", str(test_reports_dir)]
        )

        # Assert
        assert result == 0
        captured = capsys.readouterr()

        # Should show all reports from test data
        assert "d6wf-uw2w" in captured.out
        assert "lyy0-uf3p" in captured.out
        assert "ykik-ab4l" in captured.out


class TestPrintReports:
    """
    Functional tests for the "print" cli command.
    """

    @pytest.fixture(name="test_reports_dir")
    def _test_reports_dir(self) -> Path:
        """
        Fixture providing the path to test reports directory.
        """
        return Path(__file__).parent / "test_data" / "reports" / "v1"

    def test_print_basic_latest(self, capsys: CaptureFixture[str], test_reports_dir: Path) -> None:
        """
        Print the most recent iteration of a report and check for key information.
        """
        # Act
        result = entrypoint(["print", "lyy0-uf3p", "--report-directory", str(test_reports_dir)])

        # Assert
        assert result == 0
        captured = capsys.readouterr()
        assert "Memalot Report (iteration 3)" in captured.out
        assert "Possible New Leaks (iteration 3)" in captured.out
        assert "Object Type" in captured.out
        assert "Count" in captured.out
        assert "Details for __main__.MemalotObject" in captured.out
        assert "Deep size (estimated): ~1.4 KiB" in captured.out
        assert "End of Memalot Report (iteration 3)" in captured.out

    def test_print_no_leaks_message(
        self, capsys: CaptureFixture[str], test_reports_dir: Path
    ) -> None:
        """
        Print a report that contains an iteration with no leaks and check for the message.
        """
        # Act
        result = entrypoint(
            [
                "print",
                "d6wf-uw2w",
                "--num-iterations",
                "3",
                "--report-directory",
                str(test_reports_dir),
            ]
        )

        # Assert
        assert result == 0
        captured = capsys.readouterr()
        assert (
            "This is a warmup iteration. Reports will be available from the next iteration."
        ) in captured.out

    def test_print_num_iterations_two(
        self, capsys: CaptureFixture[str], test_reports_dir: Path
    ) -> None:
        """
        Print two most recent iterations and check that both are present.
        """
        # Act
        result = entrypoint(
            [
                "print",
                "lyy0-uf3p",
                "--num-iterations",
                "2",
                "--report-directory",
                str(test_reports_dir),
            ]
        )

        # Assert
        assert result == 0
        captured = capsys.readouterr()
        assert "Memalot Report (iteration 2)" in captured.out
        assert "Memalot Report (iteration 3)" in captured.out
        assert "Possible New Leaks (iteration 2)" in captured.out
        assert "Possible New Leaks (iteration 3)" in captured.out
        assert "End of Memalot Report (iteration 2)" in captured.out
        assert "End of Memalot Report (iteration 3)" in captured.out

    def test_print_filter_types(self, capsys: CaptureFixture[str], test_reports_dir: Path) -> None:
        """
        Print a report filtered by object types and verify filtering takes effect.
        """
        # Act
        result = entrypoint(
            [
                "print",
                "ykik-ab4l",
                "--report-directory",
                str(test_reports_dir),
                "--filter-types",
                "list",
            ]
        )

        # Assert
        assert result == 0
        captured = capsys.readouterr()
        assert "Possible New Leaks (iteration 3)" in captured.out
        assert "Details for __main__.MemalotObject" not in captured.out
        assert "Details for builtins.list" in captured.out

    def test_print_summary_only(self, capsys: CaptureFixture[str], test_reports_dir: Path) -> None:
        """
        Print a report with summary only and verify object details are not printed.
        """
        # Act
        result = entrypoint(
            [
                "print",
                "ykik-ab4l",
                "--report-directory",
                str(test_reports_dir),
                "--summary-only",
            ]
        )

        # Assert
        assert result == 0
        captured = capsys.readouterr()
        assert "Possible New Leaks (iteration 3)" in captured.out
        assert "Details for" not in captured.out

    def test_print_invalid_report_id(self, test_reports_dir: Path) -> None:
        """
        Attempt to print a non-existent report and expect an error.
        """
        with pytest.raises(ValueError):
            entrypoint(["print", "does-not-exist", "--report-directory", str(test_reports_dir)])



import time
from pathlib import Path
from typing import Any, Callable, Iterable, TypeVar, cast
from unittest.mock import MagicMock, create_autospec

from memalot import leak_monitor
from memalot.base import ObjectGetter

_REPORT_DIR = Path(__file__).parent / "test_data" / "reports" / "v1"


class Cache:
    def __init__(self) -> None:
        self.cache_list: list[Any] = []


class MemalotObject:
    def __init__(self, size: int) -> None:
        self.payload = [1] * size


def create_memalot_object() -> MemalotObject:
    return MemalotObject(128)


# Set save_reports=True if regenerating the test data
@leak_monitor(
    report_directory=_REPORT_DIR,
    force_terminal=True,
    check_referrers=True,
    save_reports=False,
)
def create_and_cache(cache: Cache) -> None:
    cache.cache_list.append(create_memalot_object())


def create_leaks() -> None:
    cache = Cache()
    for _ in range(3):
        create_and_cache(cache)


def create_mock(spec: Any, spec_set: bool = True, instance: bool = True) -> MagicMock:
    return cast(MagicMock, create_autospec(spec=spec, spec_set=spec_set, instance=instance))


def wait_for_assertion(
    assertion_func: Callable[[], None],
    wait_time_secs: float = 60.0,
    poll_interval_secs: float = 0.1,
) -> None:
    """
    Waits for `assertion_func` to complete without raising an `AssertionError`.

    Retries the assertion until it passes or the timeout elapses. If the timeout
    elapses, the last `AssertionError` is re-raised.
    """
    if wait_time_secs < 0:
        raise ValueError("wait_time_secs must be non-negative")
    if poll_interval_secs <= 0:
        raise ValueError("poll_interval_secs must be positive")

    deadline = time.monotonic() + wait_time_secs

    while True:
        try:
            assertion_func()
            break
        except AssertionError as error:
            if time.monotonic() >= deadline:
                raise error
            remaining = max(0.0, deadline - time.monotonic())
            time.sleep(min(poll_interval_secs, remaining))


T = TypeVar("T")


def one(iterable: Iterable[T]) -> T:
    (result,) = iterable
    return result


class FixedObjectGetter(ObjectGetter):
    """
    Gets a fixed list of objects.
    """

    def __init__(self, objects: list[Any]) -> None:
        self._objects = objects

    def get_objects(self) -> list[Any]:
        return self._objects



import gc
import weakref
from collections import Counter
from dataclasses import replace
from typing import Any, Tuple
from unittest.mock import ANY, MagicMock

import pytest
from networkx.classes import DiGraph
from pytest_mock import MockerFixture
from referrers import ReferrerGraph, ReferrerGraphNode

from memalot.base import ApproximateSize, MemalotCount
from memalot.base import ReferrerGraph as MyReferrerGraph
from memalot.memory import MemalotMemoryUsage
from memalot.monitors import FilteringObjectGetter
from memalot.options import Options
from memalot.reports import LeakSummary, TypeSummary
from memalot.snapshots import MemalotObjects, MemalotSnapshotManager, MemalotUsageSnapshot
from tests.utils_for_testing import FixedObjectGetter, create_mock

_NUM_SAME_ID_ATTEMPTS = 100


class CustomObject:
    """
    A custom object class for testing that supports weak references.
    """

    def __init__(self, value: str) -> None:
        self.value = value

    def __eq__(self, other: object) -> bool:
        return isinstance(other, CustomObject) and self.value == other.value

    def __repr__(self) -> str:
        return f"CustomObject(value='{self.value}')"


class NonWeakRefObject:
    """
    A custom object that does not support weak references.
    """

    __slots__ = ["value"]

    def __init__(self, value: str) -> None:
        self.value = value

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NonWeakRefObject) and self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)


class NonWeakRefIntObject:
    """
    A custom object containing an int that does not support weak references.
    """

    __slots__ = ["value"]

    def __init__(self, value: int) -> None:
        self.value = value

    def __eq__(self, other: object) -> bool:
        return isinstance(other, NonWeakRefIntObject) and self.value == other.value

    def __hash__(self) -> int:
        return hash(self.value)


@pytest.fixture(name="custom_objects")
def _custom_objects_fixture() -> list[CustomObject]:
    """
    Provides a list of custom objects for testing.
    """
    return [CustomObject("test1"), CustomObject("test2"), CustomObject("test3")]


@pytest.fixture(name="builtin_objects")
def _builtin_objects_fixture() -> list[object]:
    """
    Provides a list of builtin objects that don't support weak references.
    """
    return [[], {}, "string", 42, [1, 2, 3], {"key": "value"}]


@pytest.fixture(name="mixed_objects")
def _mixed_objects_fixture(
    custom_objects: list[CustomObject], builtin_objects: list[object]
) -> list[object]:
    """
    Provides a mixed list of objects with and without weak reference support.
    """
    return custom_objects + builtin_objects


@pytest.fixture(name="default_options")
def _default_options_fixture() -> Options:
    """
    Provides default Options instance for testing.
    """
    return Options()


@pytest.fixture(name="diff_objects")
def _diff_objects_fixture() -> list[object]:
    """
    Provides objects for creating MemalotUsageDiff instances.
    """
    return [
        CustomObject("diff1"),
        CustomObject("diff2"),
        [1, 2, 3],
        {"key": "value"},
        "test_string",
    ]


@pytest.fixture(name="test_memory_usage")
def _test_memory_usage_fixture() -> MemalotMemoryUsage:
    """
    Provides a MemoryUsage instance for testing.
    """
    return MemalotMemoryUsage(
        current_rss_bytes=1024000,
        peak_rss_bytes=2048000,
        system_percent_used=25.5,
        iteration_number=1,
    )


class TestMemalotUsageDiff:
    """
    Tests for the `MemalotUsageDiff` class.
    """

    def test_with_empty_list(self, default_options: Options) -> None:
        """
        Tests that MemalotUsageDiff works with an empty object list.
        """
        diff = MemalotObjects([])
        assert len(diff) == 0
        summary = diff.get_leak_summary(iteration=MemalotCount(4), options=default_options)
        assert summary == LeakSummary(
            iteration=4,
            type_summaries=[],
            max_types_in_summary=500,
        )
        object_details_list = list(diff.generate_object_details([], default_options))
        assert object_details_list == []

    @pytest.mark.parametrize("compute_size", [True, False])
    @pytest.mark.parametrize("module_prefixes", [{"override_prefix."}, None])
    def test_with_multiple_objects(
        self, default_options: Options, compute_size: bool, module_prefixes: set[str] | None
    ) -> None:
        """
        Tests that MemalotUsageDiff works with multiple objects of different types.
        """
        default_options = replace(
            default_options,
            compute_size_in_leak_summary=compute_size,
            referrers_module_prefixes=module_prefixes,
        )
        mock_referrer_graph = create_mock(spec=ReferrerGraph)
        mock_get_referrers = MagicMock(return_value=mock_referrer_graph)
        mock_networkx_graph = DiGraph()
        mock_networkx_graph.add_node(ReferrerGraphNode(name="root", id=1001, type="local"))
        mock_networkx_graph.add_node(ReferrerGraphNode(name="child", id=1002, type="attr"))
        mock_referrer_graph.to_networkx.return_value = mock_networkx_graph

        objects = [
            CustomObject("obj1"),
            CustomObject("obj2"),  # 2 CustomObjects
            [1, 2],
            [3, 4],
            [5, 6],  # 3 lists
            {"a": 1},
            {"b": 2},  # 2 dicts
            "string1",
            "string2",
            "string3",
            "string4",  # 4 strings
        ]
        diff = MemalotObjects(objects)

        summary = diff.get_leak_summary(iteration=MemalotCount(4), options=default_options)
        actuals = summary.type_summaries
        assert summary == LeakSummary(
            iteration=4,
            type_summaries=[
                TypeSummary(
                    object_type="builtins.str",
                    count=4,
                    shallow_size_bytes=actuals[0].shallow_size_bytes,  # Sizes not known
                ),
                TypeSummary(
                    object_type="builtins.list",
                    count=3,
                    shallow_size_bytes=actuals[1].shallow_size_bytes,  # Sizes not known
                ),
                TypeSummary(
                    object_type="tests.test_snapshots.CustomObject",
                    count=2,
                    shallow_size_bytes=actuals[2].shallow_size_bytes,  # Sizes not known
                ),
                TypeSummary(
                    object_type="builtins.dict",
                    count=2,
                    shallow_size_bytes=actuals[3].shallow_size_bytes,  # Sizes not known
                ),
            ],
            max_types_in_summary=500,
        )

        if compute_size:
            assert all(
                type_summary.shallow_size_bytes is not None
                for type_summary in summary.type_summaries
            )
        else:
            assert all(
                type_summary.shallow_size_bytes is None for type_summary in summary.type_summaries
            )

        object_details_list = list(
            diff.generate_object_details(
                [],
                default_options,
                get_referrers_func=mock_get_referrers,
            )
        )

        assert mock_get_referrers.call_count == 11  # One call per object
        # Check that the correct arguments were passed. We don't check every object here.
        mock_get_referrers.assert_any_call(
            CustomObject("obj1"),
            exclude_object_ids=ANY,
            max_depth=default_options.referrers_max_depth,
            max_untracked_search_depth=default_options.referrers_max_untracked_search_depth,
            timeout=default_options.referrers_search_timeout,
            single_object_referrer_limit=default_options.single_object_referrer_limit,
            module_prefixes=module_prefixes if module_prefixes is not None else {"tests."},
        )

        type_names = [object_details.object_type_name for object_details in object_details_list]
        assert set(type_names) == {
            "tests.test_snapshots.CustomObject",
            "builtins.list",
            "builtins.dict",
            "builtins.str",
        }
        object_ids = [object_details.object_id for object_details in object_details_list]
        assert set(object_ids) == {id(obj) for obj in objects}
        approx_sizes = [object_details.deep_size_bytes for object_details in object_details_list]
        # We don't know the exact sizes, but there should be 11 sizes (one per object)
        assert len(approx_sizes) == 11
        assert all(isinstance(size, ApproximateSize) for size in approx_sizes)
        referrer_graphs = [object_details.referrer_graph for object_details in object_details_list]
        for graph in referrer_graphs:
            assert isinstance(graph, MyReferrerGraph)
        # Check that each graph has the expected structure
        assert all(len(graph.graph_nodes) == 2 for graph in referrer_graphs if graph is not None)
        strings = [object_details.object_str for object_details in object_details_list]
        assert set(strings) == {
            "CustomObject(value='obj1')",
            "CustomObject(value='obj2')",
            "[1, 2]",
            "[3, 4]",
            "[5, 6]",
            "string1",
            "string2",
            "string3",
            "string4",
            "{'a': 1}",
            "{'b': 2}",
        }

    def test_ordering_by_type_count(self) -> None:
        """
        Tests that generate_object_details returns ObjectDetails instances ordered by the number
        of occurrences of each type, with the most frequent type first.
        """
        mock_referrer_graph = create_mock(spec=ReferrerGraph)
        mock_get_referrers = MagicMock(return_value=mock_referrer_graph)
        mock_networkx_graph = DiGraph()
        mock_networkx_graph.add_node(ReferrerGraphNode(name="root", id=1001, type="local"))
        mock_referrer_graph.to_networkx.return_value = mock_networkx_graph

        # Create objects with different type frequencies:
        # - 5 strings (most frequent)
        # - 3 CustomObjects (second most frequent)
        # - 2 lists (third most frequent)
        # - 1 dict (least frequent)
        objects = [
            "string1",
            "string2",
            "string3",
            "string4",
            "string5",  # 5 strings
            CustomObject("obj1"),
            CustomObject("obj2"),
            CustomObject("obj3"),  # 3 CustomObjects
            [1, 2],
            [3, 4],  # 2 lists
            {"key": "value"},  # 1 dict
        ]
        diff = MemalotObjects(objects)

        # Set max_object_details high enough to include all objects
        high_limit_options = Options(max_object_details=100)

        object_details_list = list(
            diff.generate_object_details(
                [],
                high_limit_options,
                get_referrers_func=mock_get_referrers,
            )
        )

        # Extract the type names in the order they were yielded
        type_names_in_order = [obj_details.object_type_name for obj_details in object_details_list]

        # Verify that objects are grouped by type and types appear in frequency order
        # Expected order: str (5), CustomObject (3), list (2), dict (1)
        expected_first_type = "builtins.str"  # Most frequent (5 occurrences)
        expected_second_type = "tests.test_snapshots.CustomObject"  # Second most frequent (3)
        expected_third_type = "builtins.list"  # Third most frequent (2)
        expected_fourth_type = "builtins.dict"  # Least frequent (1)

        # Find first occurrence of each type to check ordering
        str_first_index = type_names_in_order.index(expected_first_type)
        custom_obj_first_index = type_names_in_order.index(expected_second_type)
        list_first_index = type_names_in_order.index(expected_third_type)
        dict_first_index = type_names_in_order.index(expected_fourth_type)

        # Assert that types appear in the correct order (most frequent first)
        assert str_first_index < custom_obj_first_index
        assert custom_obj_first_index < list_first_index
        assert list_first_index < dict_first_index

        # Count objects by type to verify all are included
        type_counts = Counter(type_names_in_order)
        assert type_counts[expected_first_type] == 5
        assert type_counts[expected_second_type] == 3
        assert type_counts[expected_third_type] == 2
        assert type_counts[expected_fourth_type] == 1

    def test_with_max_object_details_limit(self, default_options: Options) -> None:
        """
        Tests that MemalotUsageDiff respects max_object_details and returns
        the correct subset of objects.
        """
        mock_referrer_graph = create_mock(spec=ReferrerGraph)
        mock_get_referrers = MagicMock(return_value=mock_referrer_graph)
        mock_networkx_graph = DiGraph()
        mock_networkx_graph.add_node(ReferrerGraphNode(name="root", id=1001, type="local"))
        mock_referrer_graph.to_networkx.return_value = mock_networkx_graph

        # Create many objects of different types - more than max_object_details
        objects: list[Any] = []
        # 10 CustomObjects
        for i in range(10):
            objects.append(CustomObject(f"obj{i}"))
        # 8 lists
        for i in range(8):
            objects.append([i, i + 1])
        # 6 dicts
        for i in range(6):
            objects.append({f"key{i}": f"value{i}"})
        # 4 strings
        for i in range(4):
            objects.append(f"string{i}")

        diff = MemalotObjects(objects)

        # Set max_object_details to 10 (less than total 28 objects)
        limited_options = Options(max_object_details=10)

        # With 4 types and max_object_details=10, we should get at least 2 objects per type (
        # 10//4=2) with the remainder going to the most frequent types
        object_details_list = list(
            diff.generate_object_details(
                [],
                limited_options,
                get_referrers_func=mock_get_referrers,
            )
        )

        # Should have exactly 10 objects (2 per type, plus remainder)
        assert len(object_details_list) == 10

        # Count objects by type
        type_counts = Counter(obj.object_type_name for obj in object_details_list)
        assert type_counts["tests.test_snapshots.CustomObject"] == 3
        assert type_counts["builtins.list"] == 3
        assert type_counts["builtins.dict"] == 2
        assert type_counts["builtins.str"] == 2

        # Verify that referrers were obtained for each returned object
        assert mock_get_referrers.call_count == 10

        # Verify that referrers_checked is True for all objects
        # (since check_referrers=True by default)
        for obj_details in object_details_list:
            assert obj_details.referrers_checked is True

    def test_with_check_referrers_false(self, default_options: Options) -> None:
        """
        Tests that MemalotUsageDiff works correctly when check_referrers is False.
        """
        mock_get_referrers = MagicMock()

        objects = [
            CustomObject("obj1"),
            [1, 2, 3],
            {"key": "value"},
            "test_string",
        ]
        diff = MemalotObjects(objects)

        no_referrers_options = Options(check_referrers=False)

        object_details_list = list(
            diff.generate_object_details(
                [],
                no_referrers_options,
                get_referrers_func=mock_get_referrers,
            )
        )

        # Should have all 4 objects
        assert len(object_details_list) == 4

        # Referrers function should never be called
        mock_get_referrers.assert_not_called()

        # All referrer_graph fields should be None and referrers_checked should be False
        for obj_details in object_details_list:
            assert obj_details.referrer_graph is None
            assert obj_details.referrers_checked is False

        # Other fields should still be populated correctly
        object_ids = [obj_details.object_id for obj_details in object_details_list]
        assert set(object_ids) == {id(obj) for obj in objects}

        strings = [obj_details.object_str for obj_details in object_details_list]
        assert set(strings) == {
            "CustomObject(value='obj1')",
            "[1, 2, 3]",
            "{'key': 'value'}",
            "test_string",
        }

    def test_with_object_str_exception(self, default_options: Options) -> None:
        """
        Tests that MemalotUsageDiff handles objects that raise exceptions when getting string
        representation.
        """

        class ExceptionObject:
            """
            An object that raises an exception when str() is called.
            """

            def __str__(self) -> str:
                raise ValueError("Cannot convert to string")

        mock_referrer_graph = create_mock(spec=ReferrerGraph)
        mock_get_referrers = MagicMock(return_value=mock_referrer_graph)
        mock_networkx_graph = DiGraph()
        mock_networkx_graph.add_node(ReferrerGraphNode(name="root", id=1001, type="local"))
        mock_referrer_graph.to_networkx.return_value = mock_networkx_graph

        exception_obj = ExceptionObject()
        normal_obj = CustomObject("normal")
        objects = [exception_obj, normal_obj]

        diff = MemalotObjects(objects)

        object_details_list = list(
            diff.generate_object_details(
                [],
                default_options,
                get_referrers_func=mock_get_referrers,
            )
        )

        assert len(object_details_list) == 2

        # Find the object details for the exception object
        exception_details = None
        normal_details = None
        for obj_details in object_details_list:
            if obj_details.object_id == id(exception_obj):
                exception_details = obj_details
            elif obj_details.object_id == id(normal_obj):
                normal_details = obj_details

        assert exception_details is not None
        assert normal_details is not None

        # The exception object should have an error message in its string representation
        assert "Error when getting string representation" in exception_details.object_str
        assert "Cannot convert to string" in exception_details.object_str

        # The normal object should have its normal string representation
        assert normal_details.object_str == "CustomObject(value='normal')"

    def test_with_custom_str_func(self, default_options: Options) -> None:
        """
        Tests that MemalotUsageDiff uses custom str_func when provided in options.
        """

        def custom_str_func(obj: Any, max_length: int) -> str:
            """
            A custom string function that adds a prefix.
            """
            return f"CUSTOM: {str(obj)[:max_length]}"

        mock_referrer_graph = create_mock(spec=ReferrerGraph)
        mock_get_referrers = MagicMock(return_value=mock_referrer_graph)
        mock_networkx_graph = DiGraph()
        mock_networkx_graph.add_node(ReferrerGraphNode(name="root", id=1001, type="local"))
        mock_referrer_graph.to_networkx.return_value = mock_networkx_graph

        objects = [
            CustomObject("obj1"),
            [1, 2, 3],
            {"key": "value"},
        ]
        diff = MemalotObjects(objects)

        custom_options = Options(str_func=custom_str_func)

        object_details_list = list(
            diff.generate_object_details(
                [],
                custom_options,
                get_referrers_func=mock_get_referrers,
            )
        )

        assert len(object_details_list) == 3

        # All string representations should have the custom prefix
        strings = [obj_details.object_str for obj_details in object_details_list]
        assert all(s.startswith("CUSTOM: ") for s in strings)

        # Check specific expected strings
        expected_strings = {
            "CUSTOM: CustomObject(value='obj1')",
            "CUSTOM: [1, 2, 3]",
            "CUSTOM: {'key': 'value'}",
        }
        assert set(strings) == expected_strings

    def test_eq_with_equal_objects(self) -> None:
        """
        Tests that two MemalotObjects instances with equal but different object lists are equal.
        """
        obj1 = CustomObject("test")
        obj2 = [1, 2, 3]
        objects1 = [obj1, obj2]
        objects2 = [obj1, obj2]  # Same objects, same order
        diff1 = MemalotObjects(objects1)
        diff2 = MemalotObjects(objects2)
        assert diff1 == diff2

    def test_eq_with_different_objects(self) -> None:
        """
        Tests that two MemalotObjects instances with different objects are not equal.
        """
        objects1 = [CustomObject("test1"), [1, 2, 3]]
        objects2 = [CustomObject("test2"), [4, 5, 6]]
        diff1 = MemalotObjects(objects1)
        diff2 = MemalotObjects(objects2)
        assert diff1 != diff2

    def test_eq_with_wrong_type(self) -> None:
        """
        Tests that comparing a MemalotObject instance with a string returns False.
        """
        objects1 = [CustomObject("test1"), [1, 2, 3]]
        diff1 = MemalotObjects(objects1)
        assert diff1 != "diff2"

    def test_generate_object_details_with_function_name(self, default_options: Options) -> None:
        """
        Tests that generate_object_details correctly handles function_name parameter.
        """
        # Create a mock referrer graph with nodes that have the decorator wrapper names
        mock_referrer_graph = create_mock(spec=ReferrerGraph)
        mock_get_referrers = MagicMock(return_value=mock_referrer_graph)
        mock_networkx_graph = DiGraph()

        # Add nodes with the specific names that should be replaced
        mock_networkx_graph.add_node(
            ReferrerGraphNode(
                name="memalot_decorator_inner_wrapper.memalot_decorator_inner_args",
                id=1001,
                type="local",
            )
        )
        mock_networkx_graph.add_node(
            ReferrerGraphNode(
                name="memalot_decorator_inner_wrapper.memalot_decorator_inner_kwargs",
                id=1002,
                type="local",
            )
        )
        mock_referrer_graph.to_networkx.return_value = mock_networkx_graph

        objects = [CustomObject("test")]
        diff = MemalotObjects(objects)

        object_details_list = list(
            diff.generate_object_details(
                [],
                default_options,
                function_name="my_test_function",
                get_referrers_func=mock_get_referrers,
            )
        )

        # Should have 1 object
        assert len(object_details_list) == 1
        obj_details = object_details_list[0]

        # Verify the object details
        assert obj_details.object_id == id(objects[0])

        # Verify the referrer graph has the replaced names
        assert obj_details.referrer_graph is not None
        referrer_names = [node.name for node in obj_details.referrer_graph.graph_nodes]

        # The names should have been replaced with the function name
        assert "my_test_function args" in referrer_names
        assert "my_test_function kwargs" in referrer_names

        # The original decorator wrapper names should not be present
        assert "memalot_decorator_inner_wrapper.memalot_decorator_inner_args" not in referrer_names
        assert (
            "memalot_decorator_inner_wrapper.memalot_decorator_inner_kwargs" not in referrer_names
        )


class TestMemalotUsageSnapshot:
    """
    Tests for the `MemalotUsageSnapshot` class.
    """

    def test_is_new_since_snapshot(self, default_options: Options) -> None:
        """
        Tests that new objects are correctly identified as new.
        """
        custom_objects = [CustomObject("test1"), CustomObject("test2"), CustomObject("test3")]
        # Double check that weak references can be created
        for obj in custom_objects:
            weakref.ref(obj)
        builtin_objects: list[object] = [[1, 2, 3], {"a": 1}, "string"]
        mixed_objects: list[object] = custom_objects + builtin_objects

        snapshot = MemalotUsageSnapshot(
            object_getter=FixedObjectGetter(mixed_objects),
            filter_condition=None,
        )

        # Check that existing objects are not considered new
        for mixed_obj in mixed_objects:
            assert not snapshot.is_new_since_snapshot(mixed_obj)

        # Check that new objects with the same values are considered new.
        # We can't test all of them as some are interned (like small integers and
        # strings).
        assert snapshot.is_new_since_snapshot(CustomObject("test1"))
        assert snapshot.is_new_since_snapshot([1, 2, 3])
        assert snapshot.is_new_since_snapshot({"a": 1})

        # Check that new objects with different values are considered new
        assert snapshot.is_new_since_snapshot(CustomObject("new_object"))
        assert snapshot.is_new_since_snapshot([1, 2, 3, 4])
        assert snapshot.is_new_since_snapshot({"new": "dict"})

    def test_is_in_snapshot(self, default_options: Options) -> None:
        """
        Tests that objects in the snapshot are correctly identified.
        """
        custom_objects = [CustomObject("test1"), CustomObject("test2"), CustomObject("test3")]
        # Double check that weak references can be created
        for obj in custom_objects:
            weakref.ref(obj)
        builtin_objects: list[object] = [[1, 2, 3], {"a": 1}, "string"]
        mixed_objects: list[object] = custom_objects + builtin_objects

        snapshot = MemalotUsageSnapshot(
            object_getter=FixedObjectGetter(mixed_objects),
            filter_condition=None,
        )

        # Check that existing objects are correctly identified as being in the snapshot
        for mixed_obj in mixed_objects:
            assert snapshot.is_in_snapshot(mixed_obj)

        # Check that new objects with the same values are NOT in the snapshot
        # (they are different instances)
        assert not snapshot.is_in_snapshot(CustomObject("test1"))
        assert not snapshot.is_in_snapshot([1, 2, 3])
        assert not snapshot.is_in_snapshot({"a": 1})

        # Check that new objects with different values are NOT in the snapshot
        assert not snapshot.is_in_snapshot(CustomObject("new_object"))
        assert not snapshot.is_in_snapshot([1, 2, 3, 4])
        assert not snapshot.is_in_snapshot({"new": "dict"})

    @pytest.mark.parametrize(
        "objects_in_snapshot,new_objects,expected_diff",
        [
            pytest.param(
                [],
                expected := [CustomObject("new")],
                MemalotObjects(expected),  # type: ignore[name-defined]
                id="Empty snapshot with new object",
            ),
            pytest.param(
                [CustomObject("one")],
                expected := [CustomObject("one")],
                MemalotObjects(expected),  # type: ignore[name-defined]
                id="Existing snapshot with object of same value",
            ),
            pytest.param(
                [CustomObject("old")],
                expected := [CustomObject("new"), CustomObject("new2")],
                MemalotObjects(expected),  # type: ignore[name-defined]
                id="Existing snapshot with multiple new objects",
            ),
            pytest.param(
                [CustomObject("old")],
                [],
                MemalotObjects([]),
                id="Existing snapshot with no new objects",
            ),
            pytest.param(
                [[1, 2]],
                expected_val := [[3, 4], {"c": 3}],
                MemalotObjects(expected_val),  # type: ignore[name-defined]
                id="Builtin objects",
            ),
        ],
    )
    def test_get_diff(
        self,
        objects_in_snapshot: list[object],
        new_objects: list[object],
        expected_diff: MemalotObjects,
        default_options: Options,
    ) -> None:
        """
        Tests the is_new_since_snapshot method with various object combinations.
        """
        snapshot = MemalotUsageSnapshot(
            object_getter=FixedObjectGetter(objects_in_snapshot),
            filter_condition=None,
        )
        all_objects = objects_in_snapshot + new_objects
        diff = MemalotObjects([obj for obj in all_objects if snapshot.is_new_since_snapshot(obj)])

        assert {id(obj) for obj in diff.objects} == {id(obj) for obj in expected_diff.objects}

    def test_is_new_since_snapshot_with_non_weak_ref_objects(
        self, default_options: Options
    ) -> None:
        """
        Tests that objects not supporting weak references are handled correctly in
        is_new_since_snapshot.
        """
        # Create objects that don't support weak references
        non_weak_ref_objects = [
            NonWeakRefObject("test1"),
            NonWeakRefObject("test2"),
            [1, 2, 3],
            {"key": "value"},
            "string_object",
            42,
        ]
        # Double check that weak references cannot be created
        for obj in non_weak_ref_objects:
            with pytest.raises(TypeError):
                weakref.ref(obj)

        snapshot = MemalotUsageSnapshot(
            object_getter=FixedObjectGetter(non_weak_ref_objects),
            filter_condition=None,
        )

        # Test that existing objects are not considered new
        for obj in non_weak_ref_objects:
            assert not snapshot.is_new_since_snapshot(obj)

        # Check that new objects with the same values are considered new
        # We can't test all of them as some are interned (like small integers and
        # strings).
        assert snapshot.is_new_since_snapshot(NonWeakRefObject("test1"))
        assert snapshot.is_new_since_snapshot(NonWeakRefObject("test2"))
        assert snapshot.is_new_since_snapshot([1, 2, 3])
        assert snapshot.is_new_since_snapshot({"key": "value"})

        # Test that new objects with different values are considered new
        new_non_weak_ref_obj = NonWeakRefObject("new_value")
        new_list = [4, 5, 6]
        new_dict = {"new_key": "new_value"}
        new_string = "new_string"
        new_int = 99

        assert snapshot.is_new_since_snapshot(new_non_weak_ref_obj)
        assert snapshot.is_new_since_snapshot(new_list)
        assert snapshot.is_new_since_snapshot(new_dict)
        assert snapshot.is_new_since_snapshot(new_string)
        assert snapshot.is_new_since_snapshot(new_int)

    def test_get_diff_with_non_weak_ref_objects(self, default_options: Options) -> None:
        """
        Tests that is_new_since_snapshot works correctly with objects that don't support
        weak references.
        """
        # Create initial snapshot with non-weak-ref objects
        initial_objects = [
            NonWeakRefObject("original1"),
            NonWeakRefObject("original2"),
            [1, 2],
            {"old": "dict"},
            "old_string_unique_12345",
        ]
        # Double check that weak references cannot be created
        for obj in initial_objects:
            with pytest.raises(TypeError):
                weakref.ref(obj)

        snapshot = MemalotUsageSnapshot(
            object_getter=FixedObjectGetter(initial_objects),
            filter_condition=None,
        )

        # Create new objects, some with the same value but different identities.
        new_objects = [
            NonWeakRefObject("original1"),  # Different object ID; will be in diff
            NonWeakRefObject("new_value"),  # New value; will be in diff
            [1, 2],  # Different object ID; will be in diff
            [3, 4],  # New value; will be in diff
            {"old": "dict"},  # Different object ID; will be in diff
            {"new": "dict"},  # New value; will be in diff
            "new_string_unique_67890",  # New value; will be in diff
        ]

        all_objects = initial_objects + new_objects
        diff = MemalotObjects([obj for obj in all_objects if snapshot.is_new_since_snapshot(obj)])

        assert {id(obj) for obj in diff.objects} == {id(obj) for obj in new_objects}

    def test_weak_ref_object_gced(self, default_options: Options) -> None:
        """
        Tests that if an object supporting weak references is garbage collected,
        it is no longer considered part of the snapshot.
        """
        snapshot, gced_object_id, non_gced_object = _generate_snapshot_with_weak_ref_object(
            default_options=default_options
        )
        gc.collect()
        object_ids = snapshot.active_object_ids
        assert gced_object_id not in object_ids
        assert id(non_gced_object) in object_ids

    def test_non_weak_ref_object_id_reused(self, default_options: Options) -> None:
        """
        Tests that if an object ID is reused for a non-weak-ref object, then the objects are still
        considered different if they have different hashes.
        """
        snapshot, previous_object_id = self._get_snapshot_with_non_weak_ref_object1(
            default_options=default_options
        )
        # Try (multiple times) to create a new object with the same ID as the previous object.
        # We skip the test if it doesn't happen. This isn't ideal but is probably better than not
        # testing this case at all.
        for _ in range(_NUM_SAME_ID_ATTEMPTS):
            new_obj = self._get_non_weak_ref_object2()
            new_obj_id = id(new_obj)
            # Check that the new object has the same ID as the previous object.
            # This behaviour is not guaranteed, as the object ID may not be reused
            # (this is implementation-dependent).
            if new_obj_id == previous_object_id:
                break
        else:
            pytest.skip(
                f"Could not create object with same ID as previous object after "
                f"{_NUM_SAME_ID_ATTEMPTS} attempts - skipping test"
            )
        # Even though the object ID is the same, it is a different object (it has
        # a different hash), so should be considered new.
        # Note: this wouldn't work properly if the object hash was the same, even
        # if the object was new. This is a current limitation of the implementation.
        assert snapshot.is_new_since_snapshot(new_obj)

    def test_excludes_internal_machinery_objects(self, default_options: Options) -> None:
        """
        Tests that objects that are created as part of creating the snapshot
        are excluded from the snapshot.
        """
        original_objects = [
            CustomObject("internal_test1"),  # Weak-ref object
            CustomObject("internal_test2"),  # Weak-ref object
            [1, 2, 3],  # Non-weak-ref object
        ]

        # A mock version of the internal machinery objects
        internal_machinery_objects = [
            CustomObject("machinery1"),  # Weak-ref object
            CustomObject("machinery2"),  # Weak-ref object
            [7, 8, 9],  # Non-weak-ref object
        ]

        get_objects_func = MagicMock(side_effect=[original_objects, internal_machinery_objects])

        snapshot = MemalotUsageSnapshot(
            object_getter=FilteringObjectGetter(
                get_objects_func=get_objects_func,
                options=default_options,
                snapshot_managers=[create_mock(MemalotSnapshotManager)],
            ),
            filter_condition=None,
        )

        # The original objects should not be considered new
        for obj in original_objects:
            assert not snapshot.is_new_since_snapshot(obj)

        # The (mock) internal machinery objects should not be considered new
        for obj in internal_machinery_objects:
            assert not snapshot.is_new_since_snapshot(obj)

        # A new object created after get_objects_func is called is considered new
        new_normal_object = CustomObject("new_normal")
        assert snapshot.is_new_since_snapshot(new_normal_object)

    def test_filter_condition_filters_objects(self, default_options: Options) -> None:
        """
        Tests that the filter_condition parameter properly filters objects during snapshot creation.
        """
        all_objects = [
            CustomObject("obj1"),
            CustomObject("obj2"),
            CustomObject("obj3"),
            CustomObject("obj4"),
        ]

        # Create a filter that only allows objects with "2" or "4" in their value
        def filter_func(obj: object) -> bool:
            if isinstance(obj, CustomObject):
                return "2" in obj.value or "4" in obj.value
            return False

        snapshot = MemalotUsageSnapshot(
            object_getter=FixedObjectGetter(all_objects),
            filter_condition=filter_func,
        )

        # Only obj2 and obj4 should be in the snapshot
        assert snapshot.is_in_snapshot(all_objects[1])  # obj2
        assert snapshot.is_in_snapshot(all_objects[3])  # obj4
        assert not snapshot.is_in_snapshot(all_objects[0])  # obj1
        assert not snapshot.is_in_snapshot(all_objects[2])  # obj3

        # Verify that the snapshot only contains 2 objects
        assert len(snapshot.active_object_ids) == 2

    def _get_non_weak_ref_object1(self) -> NonWeakRefIntObject:
        return NonWeakRefIntObject(1)

    def _get_non_weak_ref_object2(self) -> NonWeakRefIntObject:
        return NonWeakRefIntObject(2)

    def _get_snapshot_with_non_weak_ref_object1(
        self, default_options: Options
    ) -> Tuple[MemalotUsageSnapshot, int]:
        obj = self._get_non_weak_ref_object1()
        snapshot = MemalotUsageSnapshot(
            object_getter=FixedObjectGetter([obj]), filter_condition=None
        )
        return snapshot, id(obj)


class TestMemalotSnapshotManager:
    """
    Tests for the `MemalotSnapshotManager` class.
    """

    def test_generate_new_snapshot(self, default_options: Options) -> None:
        """
        Tests that generate_new_snapshot creates multiple snapshots correctly from a
        new manager instance.
        """
        manager = MemalotSnapshotManager(
            report_id="test-snapshot-123",
        )

        # Initial state - no snapshot
        assert manager.most_recent_snapshot is None
        assert manager.most_recent_snapshot_time is None

        # Create first snapshot
        objects1 = [CustomObject("snap1"), [1, 2], {"key": "value1"}]
        manager.generate_new_snapshot(
            object_getter=FixedObjectGetter(objects1), since_snapshot=None
        )

        # Verify first snapshot contains the expected object IDs
        first_snapshot = manager.most_recent_snapshot
        assert first_snapshot is not None
        assert manager.most_recent_snapshot_time is not None
        first_time = manager.most_recent_snapshot_time
        expected_ids1 = {id(obj) for obj in objects1}
        assert first_snapshot.active_object_ids == expected_ids1

        # Create second snapshot - should replace the first
        objects2 = [CustomObject("snap2"), [3, 4], {"key": "value2"}, "new_string"]
        manager.generate_new_snapshot(
            object_getter=FixedObjectGetter(objects2), since_snapshot=first_snapshot
        )

        # Verify second snapshot contains the expected object IDs and replaced the first
        second_snapshot = manager.most_recent_snapshot
        assert second_snapshot is not None
        assert manager.most_recent_snapshot_time is not None
        assert second_snapshot != first_snapshot
        assert manager.most_recent_snapshot_time >= first_time
        expected_ids2 = {id(obj) for obj in objects2}
        assert second_snapshot.active_object_ids == expected_ids2

        # Create third snapshot to ensure multiple rotations work
        objects3 = [CustomObject("snap3"), NonWeakRefObject("test")]
        manager.generate_new_snapshot(
            object_getter=FixedObjectGetter(objects3), since_snapshot=second_snapshot
        )

        # Verify third snapshot contains the expected object IDs
        third_snapshot = manager.most_recent_snapshot
        assert third_snapshot is not None
        assert manager.most_recent_snapshot_time is not None
        expected_ids3 = {id(obj) for obj in objects3}
        assert third_snapshot.active_object_ids == expected_ids3

    def test_clear_snapshots(self, default_options: Options) -> None:
        """
        Tests that clear_snapshots removes all snapshot data.
        """
        manager = MemalotSnapshotManager(
            report_id="test-clear-123",
        )

        # Create a snapshot first
        objects = [CustomObject("clear_test"), [1, 2, 3]]
        manager.generate_new_snapshot(object_getter=FixedObjectGetter(objects), since_snapshot=None)

        # Verify snapshot exists
        assert manager.most_recent_snapshot is not None
        assert manager.most_recent_snapshot_time is not None

        # Clear snapshots
        manager.clear_snapshots()

        # Verify all snapshot data is cleared
        assert manager.most_recent_snapshot is None
        assert manager.most_recent_snapshot_time is None

    def test_rotate_memory_usage(self, default_options: Options) -> None:
        """
        Tests that rotate_memory_usage tracks memory usage across multiple calls from a new
        manager instance.
        """
        manager = MemalotSnapshotManager(
            report_id="test-memory-123",
        )

        # Initial state - no memory usage tracked
        assert manager.most_recent_reported_usage is None

        # First rotation - should return None for old usage
        old_usage1, new_usage1 = manager.rotate_memory_usage(iteration=MemalotCount(1))

        assert old_usage1 is None
        assert new_usage1 is not None
        assert new_usage1.iteration_number == 1
        assert manager.most_recent_reported_usage == new_usage1

        # Second rotation - should return previous usage as old
        old_usage2, new_usage2 = manager.rotate_memory_usage(iteration=MemalotCount(2))

        assert old_usage2 == new_usage1
        assert new_usage2 is not None
        assert new_usage2.iteration_number == 2
        assert manager.most_recent_reported_usage == new_usage2

        # Third rotation - ensure multiple rotations work
        old_usage3, new_usage3 = manager.rotate_memory_usage(iteration=MemalotCount(3))

        assert old_usage3 == new_usage2
        assert new_usage3 is not None
        assert new_usage3.iteration_number == 3
        assert manager.most_recent_reported_usage == new_usage3

    def test_report_id_property(self, default_options: Options) -> None:
        """
        Tests that the report_id property returns the correct report ID.
        """
        report_id = "test-property-id-456"
        manager = MemalotSnapshotManager(
            report_id=report_id,
        )

        assert manager.report_id == report_id

    def test_snapshot_created_with_correct_parameters(self, mocker: MockerFixture) -> None:
        """
        Tests that MemalotUsageSnapshot is created with filter_condition=None when
        since_snapshot=None.
        """
        mock_snapshot_init = mocker.patch(
            "memalot.snapshots.MemalotUsageSnapshot.__init__", return_value=None
        )
        mock_get_objects_func = MagicMock()
        manager = MemalotSnapshotManager(
            report_id="report_id",
        )
        manager.generate_new_snapshot(object_getter=mock_get_objects_func, since_snapshot=None)
        mock_snapshot_init.assert_called_once_with(
            object_getter=mock_get_objects_func,
            filter_condition=None,
        )

    def test_snapshot_created_with_filter_condition_when_new_objects_only(
        self, mocker: MockerFixture, default_options: Options
    ) -> None:
        """
        Tests that MemalotUsageSnapshot is created with a filter_condition when since_snapshot
        is provided. The filter_condition should be the is_new_since_snapshot method of the
        since_snapshot.
        """
        mock_snapshot_init = mocker.patch(
            "memalot.snapshots.MemalotUsageSnapshot.__init__", return_value=None
        )
        mock_get_objects_func = MagicMock()
        manager = MemalotSnapshotManager(
            report_id="report_id",
        )
        manager.generate_new_snapshot(object_getter=mock_get_objects_func, since_snapshot=None)

        first_snapshot = manager.most_recent_snapshot

        manager.generate_new_snapshot(
            object_getter=mock_get_objects_func, since_snapshot=first_snapshot
        )

        # The second call should have a filter_condition that is the is_new_since_snapshot
        # method of the first snapshot
        assert mock_snapshot_init.call_count == 2
        second_call = mock_snapshot_init.call_args_list[1]
        assert second_call.kwargs["object_getter"] is mock_get_objects_func
        assert first_snapshot is not None  # Keep Mypy happy
        assert second_call.kwargs["filter_condition"] == first_snapshot.is_new_since_snapshot

    def test_new_objects_only_requires_existing_snapshot(self, default_options: Options) -> None:
        """
        Tests that generate_new_snapshot with since_snapshot parameter is correctly applied.
        """
        manager = MemalotSnapshotManager(
            report_id="report_id",
        )

        # This should work fine - no error expected when since_snapshot is None
        manager.generate_new_snapshot(
            object_getter=FixedObjectGetter([]),
            since_snapshot=None,
        )


def _generate_snapshot_with_weak_ref_object(
    default_options: Options,
) -> Tuple[MemalotUsageSnapshot, int, CustomObject]:
    """
    Helper to generate two weak-referencable objects and return a snapshot containing them.

    When this function returns, the first object will be eligible for garbage collection, but
    the second will not (since it's returned).
    """
    gced_object = CustomObject("gced_object")
    non_gced_object = CustomObject("non_gced_object")
    # Double check that weak references can be created
    weakref.ref(gced_object)
    weakref.ref(non_gced_object)
    snapshot = MemalotUsageSnapshot(
        object_getter=FixedObjectGetter([gced_object, non_gced_object]),
        filter_condition=None,
    )
    return snapshot, id(gced_object), non_gced_object



"""
Integration test for the converging tree example.
"""

from __future__ import annotations

from pathlib import Path
from typing import Optional

import pytest

from memalot import leak_monitor
from tests_integration.utils import (
    assert_iteration_count,
    assert_object_details_types,
    assert_object_type_details,
    assert_referrer_graphs_have_root_names,
    assert_type_summaries,
    assert_warmup_iteration,
    read_single_report,
)


class TreeNode:
    """
    A node in a tree with a parent reference.
    """

    def __init__(self, parent: Optional["TreeNode"]):
        self.parent = parent


class RootNode(TreeNode):
    """
    A root node with no parent.
    """

    pass


class TreeNodeHolder:
    """
    A holder for multiple tree nodes (leaf nodes).
    """

    def __init__(self, node1: TreeNode, node2: TreeNode, node3: TreeNode, node4: TreeNode):
        self.node1 = node1
        self.node2 = node2
        self.node3 = node3
        self.node4 = node4


@pytest.mark.integration
class TestConvergingTreeExample:
    """
    Tests for the converging tree example.
    """

    def test_converging_tree_example(self, tmp_path: Path) -> None:
        """
        Tests that the converging tree example generates the expected report content.

        This test verifies that:
        - TreeNodes created in a converging tree structure are properly detected as leaks
        - Each iteration reports exactly 6 TreeNodes (1 root, 2 level1, 4 level2)
          and 1 TreeNodeHolder
        - Type summaries contain RootNode, TreeNode, and TreeNodeHolder types
        - Referrer graphs correctly identify roots and leaf_holders as roots
        """

        # given: a construct_converging_tree function with leak monitoring
        @leak_monitor(
            report_directory=tmp_path,
            max_object_details=4,
        )
        def construct_converging_tree(
            roots: list[RootNode], leaf_holders: list[TreeNodeHolder]
        ) -> None:
            root = RootNode(parent=None)
            level1_1 = TreeNode(parent=root)
            level1_2 = TreeNode(parent=root)
            level2_1 = TreeNode(parent=level1_1)
            level2_2 = TreeNode(parent=level1_1)
            level2_3 = TreeNode(parent=level1_2)
            level2_4 = TreeNode(parent=level1_2)
            roots.append(root)
            leaf_holder = TreeNodeHolder(
                node1=level2_1, node2=level2_2, node3=level2_3, node4=level2_4
            )
            leaf_holders.append(leaf_holder)

        roots: list[RootNode] = []
        leaf_holders: list[TreeNodeHolder] = []

        # when: we call construct_converging_tree 3 times to generate reports
        for _ in range(3):
            construct_converging_tree(roots, leaf_holders)

        # then: read the report from disk and verify
        full_report = read_single_report(tmp_path=tmp_path)
        assert_iteration_count(full_report=full_report, expected_count=3)

        # Iteration 1: warmup - no leaks reported
        iteration_1 = full_report.iterations[0]
        assert iteration_1.iteration_number == 1
        assert_warmup_iteration(iteration=iteration_1)

        # Iteration 2: exactly 1 RootNode, 4 TreeNodes, and 1 TreeNodeHolder
        iteration_2 = full_report.iterations[1]
        assert iteration_2.iteration_number == 2

        # Check type summaries
        type_summaries = iteration_2.leak_summary.type_summaries
        expected_types = {
            "tests_integration.test_converging_tree.RootNode",
            "tests_integration.test_converging_tree.TreeNode",
            "tests_integration.test_converging_tree.TreeNodeHolder",
        }
        # We expect 1 RootNode, 6 TreeNodes (2 level1 + 4 level2), 1 TreeNodeHolder
        assert_type_summaries(
            type_summaries=type_summaries, expected_types=expected_types, expected_counts={1, 6}
        )

        # Check object details
        object_details_list = list(iteration_2.object_details_list)
        assert_object_details_types(
            object_details_list=object_details_list, expected_types=expected_types
        )

        # Verify RootNode objects
        assert_object_type_details(
            object_details_list=object_details_list,
            object_type="tests_integration.test_converging_tree.RootNode",
            expected_count=1,
            expected_target_names={"RootNode (object)"},
        )

        # Verify TreeNode objects (sample of 2, since max_object_details=4)
        assert_object_type_details(
            object_details_list=object_details_list,
            object_type="tests_integration.test_converging_tree.TreeNode",
            expected_count=2,
            expected_target_names={"TreeNode (object)"},
        )

        # Verify TreeNodeHolder objects
        assert_object_type_details(
            object_details_list=object_details_list,
            object_type="tests_integration.test_converging_tree.TreeNodeHolder",
            expected_count=1,
            expected_target_names={"TreeNodeHolder (object)"},
        )

        # Verify root names for all objects
        expected_root_names = {
            "test_converging_tree_example.roots (local)",
            "test_converging_tree_example.leaf_holders (local)",
            "construct_converging_tree args (local)",
        }
        assert_referrer_graphs_have_root_names(
            object_details_list=object_details_list,
            expected_root_names=expected_root_names,
        )

        # Iteration 3: same as iteration 2
        iteration_3 = full_report.iterations[2]
        assert iteration_3.iteration_number == 3

        type_summaries_3 = iteration_3.leak_summary.type_summaries
        assert_type_summaries(
            type_summaries=type_summaries_3, expected_types=expected_types, expected_counts={1, 6}
        )



import pytest

from memalot.options import InvalidOptionsError, Options


class TestOptions:
    def test_invalid_options(self) -> None:
        with pytest.raises(InvalidOptionsError):
            Options(output_func=lambda x: None, tee_console=False, force_terminal=True)



import json
import random
import string
import sys
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from pathlib import Path
from typing import Generator, List

from pydantic import TypeAdapter, field_serializer
from pydantic.dataclasses import dataclass
from rich import box
from rich.console import NewLine as RichNewLine
from rich.console import RenderableType
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from memalot import networkx_copy
from memalot.base import ApproximateSize, CachingIterable, ReferrerGraph
from memalot.memory import MemalotMemoryUsage
from memalot.options import Options
from memalot.output import NewBlock, Output
from memalot.themes import (
    COUNT,
    END_OF_SECTION_TEXT,
    HEADING_BORDER,
    MEMORY_DECREASE,
    MEMORY_INCREASE,
    OBJECT_DETAILS_TITLE,
    SIZE_TITLE,
    STRING_REPR_TITLE,
    TABLE_BORDER,
    TABLE_HEADER,
    TABLE_TITLE,
    TITLE,
)
from memalot.utils import as_mib, as_mib_sf, convert_graph_nodes_to_printable, format_bytes

_CURRENT_FILE_VERSION = 1
"""
The current file format version. Used when writing files.
"""

_REPORT_DIR_PREFIX = "memalot_report_"
"""
The prefix of the report directory name.
"""

_ITERATION_REPORT_PREFIX = "iteration_report_"
"""
The prefix of the iteration report file name.
"""

_SUMMARY_FILE_NAME = "summary.json"
"""
The name of the summary file.
"""


@dataclass
class ReportMetadata:
    """
    Metadata relating to a single report. This is generated at the start of execution.
    """

    report_id: str
    """
    A unique ID of the report.
    """
    entrypoint: str
    """
    The entrypoint of the report. This is the main function that was used to start the
    program that was analyzed.
    """
    arguments: List[str]
    """
    The arguments that were used to start the program that was analyzed.
    """
    start_time: datetime
    """
    The start time of the report.
    """


@dataclass
class ReportSummary:
    """
    A summary of a single report, generated at the end of execution.
    """

    metadata: ReportMetadata
    """
    Metadata about the report.
    """

    iteration_count: int
    """
    The number of iterations that were run. This can be used to, for example, get the most
    recent iteration of the report, since iteration numbers are sequential.
    """


@dataclass
class MemoryUsageOutput(Output):
    """
    An `Output` object representing the memory usage.
    """

    current_rss_bytes: int
    """
    The current resident set size (RSS) in bytes.
    """

    system_percent_used: float
    """
    The percent of total physical system memory used.
    """

    peak_rss_bytes: int | None = None
    """
    The peak resident set size (RSS) in bytes. This is `None` if not supported on the
    current platform.
    """

    current_rss_diff: int | None = None
    """
    The difference in RSS between the current and previous usage. This is `None` if
    previous usage is not provided.
    """

    peak_rss_diff: int | None = None
    """
    The difference in RSS between the peak and previous peak. This is `None` if
    previous usage is not provided.
    """

    diff_from_iteration: int | None = None
    """
    The iteration number that the difference was calculated from. This is `None` if
    previous usage is not provided.
    """

    def to_renderables(self) -> list[RenderableType]:
        parts = [f"[{TITLE}]Memory used[/{TITLE}]: {as_mib(self.current_rss_bytes)}"]
        if self.current_rss_diff is not None:
            sign = "+" if self.current_rss_diff >= 0 else ""
            tag = MEMORY_INCREASE if self.current_rss_diff > 0 else MEMORY_DECREASE
            parts.append(f" ([{tag}]{sign}[/{tag}]{as_mib_sf(self.current_rss_diff, tag=tag)})")
        parts.append(f" / {self.system_percent_used:.2f}% sys")
        if self.peak_rss_bytes is not None:
            parts.append("\n")
            parts.append(f"[{TITLE}]Peak memory[/{TITLE}]: {as_mib(self.peak_rss_bytes)}")
            if self.peak_rss_diff is not None:
                sign = "+" if self.peak_rss_diff >= 0 else ""
                tag = MEMORY_INCREASE if self.peak_rss_diff > 0 else MEMORY_DECREASE
                parts.append(f" ([{tag}]{sign}[/{tag}]{as_mib_sf(self.peak_rss_diff, tag=tag)})")
        return ["".join(parts)]


@dataclass
class TypeSummary:
    """
    A summary of a specific type of object.
    """

    object_type: str
    """
    The name of the type of object that was found to be leaking.
    """
    count: int
    """
    The number of objects of this type that were found to be leaking.
    """
    shallow_size_bytes: ApproximateSize | None = None
    """
    The total shallow size of all objects of this type that were found to be leaking.

    If the total size could not be determined, then the `upper_bound_known` field of
    `deep_size_bytes` will be `False`.

    This may be None if the shallow size was not computed.
    """


@dataclass
class LeakSummary(Output):
    """
    A summary of potential leaks.
    """

    iteration: int
    """
    The iteration that this summary is for
    """

    type_summaries: list[TypeSummary]
    """
    A list of summaries for each type of object that was found to be leaking.
    """

    max_types_in_summary: int | None = None
    """
    The maximum number of types in the summary. This may be `None` in older versions.
    """

    def to_renderables(self) -> list[RenderableType]:
        table = Table(
            title=f"Possible New Leaks (iteration {self.iteration})",
            box=box.ROUNDED,
            header_style=TABLE_HEADER,
            title_style=TABLE_TITLE,
            border_style=TABLE_BORDER,
        )
        table.add_column("Object Type", justify="left", no_wrap=False, overflow="fold")
        table.add_column("Count", justify="right", no_wrap=True)
        show_sizes = any(summary.shallow_size_bytes is not None for summary in self.type_summaries)
        if show_sizes:
            table.add_column("Shallow size (estimated)", justify="right", no_wrap=True)
        for summary in self.type_summaries:
            if summary.shallow_size_bytes is not None:
                byte_string = format_bytes(summary.shallow_size_bytes.approx_size)
                formatted_bytes = f"{summary.shallow_size_bytes.prefix}{byte_string}"
            else:
                formatted_bytes = "Unknown"
            row_data = [
                summary.object_type,
                f"[{COUNT}]{summary.count}[/{COUNT}]",
            ]
            if show_sizes:
                row_data.append(formatted_bytes)
            table.add_row(*row_data)
        renderables: list[RenderableType] = [table]
        if (
            self.max_types_in_summary is not None
            and len(self.type_summaries) >= self.max_types_in_summary
        ):
            renderables.append(f"(truncated to {self.max_types_in_summary} entries)")
        return renderables


@dataclass
class ObjectDetails(Output):
    """
    Represents details of a specific object, including its referrers.
    """

    object_type_name: str
    object_id: int
    object_str: str
    deep_size_bytes: ApproximateSize
    referrer_graph: ReferrerGraph | None = None
    referrers_checked: bool = True

    def to_renderables(self) -> list[RenderableType]:
        title = (
            f"[{OBJECT_DETAILS_TITLE}]Details for {self.object_type_name} (id="
            f"{self.object_id})[/{OBJECT_DETAILS_TITLE}]"
        )
        size_str = (
            f"[{OBJECT_DETAILS_TITLE}]{self.deep_size_bytes.prefix}[/{OBJECT_DETAILS_TITLE}]"
            f"{format_bytes(self.deep_size_bytes.approx_size, tag=OBJECT_DETAILS_TITLE)}"
        )
        size = f"[{SIZE_TITLE}]Deep size (estimated):[/{SIZE_TITLE}] {size_str}"
        if self.referrer_graph is None or len(self.referrer_graph) == 0:
            if self.referrers_checked:
                referrer_graph = "No referrers found"
            else:
                referrer_graph = (
                    "Not checking referrers. Set the check_referrers option to "
                    "True to turn on referrer checking"
                )
        else:
            # We avoid using Rich's tree here for now, as it seems to have issues
            # with wrapping and cropping. See https://github.com/Textualize/rich/issues/3785.
            printable_graph = convert_graph_nodes_to_printable(self.referrer_graph)
            network_text = networkx_copy.generate_network_text(
                printable_graph, override_glyphs=networkx_copy.UtfUndirectedGlyphs
            )  # type: ignore
            referrer_graph = "\n" + "\n".join(line for line in network_text)
        return [
            Panel(
                title,
                box=box.ASCII,
                expand=False,
                padding=0,
                border_style=HEADING_BORDER,
            ),
            RichNewLine(),
            size,
            RichNewLine(),
            referrer_graph,
            RichNewLine(),
            f"[{STRING_REPR_TITLE}]String representation:[/{STRING_REPR_TITLE}]",
            self.object_str,
        ]

    def __eq__(self, other: object) -> bool:
        # noinspection PyProtocol
        assert isinstance(other, ObjectDetails)
        # noinspection PyUnresolvedReferences
        return (
            self.object_type_name == other.object_type_name
            and self.object_id == other.object_id
            and self.deep_size_bytes == other.deep_size_bytes
            and self.referrers_checked == other.referrers_checked
            and self.referrer_graph == other.referrer_graph
        )

    def __hash__(self) -> int:
        raise NotImplementedError("Hashing ObjectDetails is not supported")


@dataclass(config={"arbitrary_types_allowed": True})
class ReportIteration(Output):
    """
    A single iteration of a report.
    """

    report_id: str
    """
    The ID of the report that this iteration belongs to.
    """
    iteration_number: int
    """
    The iteration number of the report.
    """
    start_time: datetime
    """
    The start time of the iteration.
    """
    end_time: datetime
    """
    The end time of the iteration.
    """
    memory_usage: MemoryUsageOutput
    """
    The memory usage of the iteration.
    """
    leak_summary: LeakSummary
    """
    A summary of the leaks found in this iteration.
    """
    object_details_list: CachingIterable[ObjectDetails] | list[ObjectDetails]
    """
    A list of object details for each potential leak.
    """

    def to_renderables(self) -> Generator[RenderableType, None, None]:
        """
        Converts this report iteration to a generator of rich renderable objects.

        This yields all the individual renderable elements that make up the report
        for this iteration, in the order they should be displayed.
        """
        yield RichNewLine()
        yield RichNewLine()

        report_id_text = f"Report ID: {self.report_id}"

        panel_text = Text.from_markup(
            report_id_text + "\n" + str(self.memory_usage.to_renderables()[0]), justify="center"
        )
        yield Panel(
            panel_text,
            expand=False,
            padding=1,
            border_style=HEADING_BORDER,
            title=f"[bold]Memalot Report[/bold] (iteration {self.iteration_number})",
            title_align="center",
        )

        yield NewBlock("")

        yield RichNewLine()

        if self.leak_summary.type_summaries:
            for renderable in self.leak_summary.to_renderables():
                yield renderable
            yield RichNewLine()
        else:
            if self.iteration_number == 1:
                yield (
                    "This is a warmup iteration. Reports will be available from the next iteration."
                )
            else:
                yield "No leaks found during this iteration."
            yield RichNewLine()

        yield NewBlock("")

        if self.leak_summary.type_summaries:
            yield "Memalot is generating object details. This may take some time..."
            yield NewBlock("")

        for object_details in self.object_details_list:
            for renderable in object_details.to_renderables():
                yield renderable
            yield RichNewLine()
            yield NewBlock("")

        yield (
            f"[{END_OF_SECTION_TEXT}]End of Memalot Report "
            f"(iteration {self.iteration_number})[/{END_OF_SECTION_TEXT}]"
        )

    @field_serializer("object_details_list")
    def serialize_referrer_graph(
        self, referrers: CachingIterable[ObjectDetails] | list[ObjectDetails]
    ) -> list[ObjectDetails]:
        if isinstance(referrers, CachingIterable):
            return list(referrers)
        else:
            return referrers


@dataclass
class FullReport:
    """
    A full report.
    """

    summary: ReportSummary
    """
    A summary of the report.
    """
    iterations: list[ReportIteration]
    """
    A list of all iterations of the report. Each iteration contains information about the
    memory usage and potential leaks at that point in time.
    """


class ReportWriter(ABC):
    """
    A writer that writes report data to a sink. This is typically a file.
    """

    @abstractmethod
    def write_iteration(self, iteration: ReportIteration) -> None:  # pragma: no cover
        pass

    @property
    @abstractmethod
    def report_id(self) -> str:  # pragma: no cover
        pass


class ReportReader(ABC):
    """
    A reader that reads report data from a source. This is typically a file.
    """

    @abstractmethod
    def get_report_summaries(self) -> list[ReportSummary]:
        """
        Gets the summaries of all reports in the given directory. If no directory is
        provided, the default report root is used.

        If the directory does not exist, an empty list is returned.
        """
        pass  # pragma: no cover

    @abstractmethod
    def get_full_report(self, report_id: str, num_iterations: int) -> FullReport:
        """
        Gets the full report for a given report ID. The number of iterations to return
        is specified by the `num_iterations` parameter. The most recent `num_iterations`
        are returned, ordered by iteration number.

        If the report does not exist, a `ValueError` is raised.
        """
        pass  # pragma: no cover


class NoOpReportWriter(ReportWriter):
    """
    A writer that does nothing.
    """

    def __init__(self, report_summary: ReportMetadata) -> None:
        self._report_id = report_summary.report_id

    def write_iteration(self, iteration: ReportIteration) -> None:
        # Do nothing
        pass

    @property
    def report_id(self) -> str:
        return self._report_id


class FileReportWriter(ReportWriter):
    """
    A writer that writes report data to a file within the report directory. Each
    iteration is written to a separate file suffixed with the iteration number.

    The writer is initialized with the path to the report directory and the report
    summary, which is written to a file named `summary.json`.
    """

    def __init__(self, report_directory: Path, report_summary: ReportMetadata) -> None:
        self._report_directory = report_directory
        self._report_summary = report_summary
        summary_file = self._report_directory / _SUMMARY_FILE_NAME
        with open(summary_file, "wb") as f:
            f.write(TypeAdapter(type(self._report_summary)).dump_json(self._report_summary))
        version_file = self._report_directory / "version"
        with open(version_file, "w") as f:
            f.write(str(_CURRENT_FILE_VERSION))

    def write_iteration(self, iteration: ReportIteration) -> None:
        iteration_file = (
            self._report_directory / f"{_ITERATION_REPORT_PREFIX}{iteration.iteration_number}.json"
        )
        with open(iteration_file, "wb") as f:
            f.write(TypeAdapter(type(iteration)).dump_json(iteration))

    @property
    def report_id(self) -> str:
        return self._report_summary.report_id


class FileReportReader(ReportReader):
    def __init__(self, report_directory: Path | None = None) -> None:
        self._report_root = _get_report_root(
            override=str(report_directory) if report_directory else None
        )

    def get_report_summaries(self) -> List[ReportSummary]:
        """
        Gets the summaries of all reports in the given directory. If no directory is
        provided, the default report root is used.

        If the directory does not exist, an empty list is returned.
        """
        if not self._report_root.exists():
            return []
        else:
            summary_list = []
            # This is not a list comprehension because otherwise it shows up on leak reports
            # in Python versions < 3.12.
            report_dirs = []
            for item in self._report_root.iterdir():
                if item.is_dir() and item.name.startswith(_REPORT_DIR_PREFIX):
                    report_dirs.append(item)
            for report_dir in report_dirs:
                summary_path = report_dir / _SUMMARY_FILE_NAME
                if summary_path.exists():
                    with open(summary_path, "r") as f:
                        data = json.load(f)
                        summary_start: ReportMetadata = TypeAdapter(ReportMetadata).validate_python(
                            data
                        )
                    iteration_files = list(report_dir.glob(f"{_ITERATION_REPORT_PREFIX}*.json"))
                    iteration_count = len(iteration_files)
                    summary = ReportSummary(
                        metadata=summary_start,
                        iteration_count=iteration_count,
                    )
                    summary_list.append(summary)
            return summary_list

    def get_full_report(self, report_id: str, num_iterations: int) -> FullReport:
        report_dir = self._report_root / f"{_REPORT_DIR_PREFIX}{report_id}"
        if not report_dir.exists():
            raise ValueError(f"Report with ID {report_id} not found.")

        summary_path = report_dir / _SUMMARY_FILE_NAME
        with open(summary_path, "r") as f:
            data = json.load(f)
            summary_start = TypeAdapter(ReportMetadata).validate_python(data)

        iteration_files = sorted(
            report_dir.glob(f"{_ITERATION_REPORT_PREFIX}*.json"),
            key=lambda p: int(p.stem.split(f"{_ITERATION_REPORT_PREFIX}")[1]),
            reverse=True,
        )

        iterations = []
        for iteration_file in iteration_files[:num_iterations]:
            with open(iteration_file, "r") as f:
                data = json.load(f)
                iteration = TypeAdapter(ReportIteration).validate_python(data)
                iterations.append(iteration)

        iterations.sort(key=lambda i: i.iteration_number)

        summary = ReportSummary(
            metadata=summary_start,
            iteration_count=len(iteration_files),
        )

        return FullReport(summary=summary, iterations=iterations)


def get_report_writer(options: Options) -> ReportWriter:
    """
    Gets an report writer for the given options.

    If the report directory is not specified, the default report root is used.
    The report root is a directory in the user's home directory called `.memalot/reports`.
    """
    start_time = datetime.now(timezone.utc)
    report_id = _generate_report_id(start_time)

    report_summary = ReportMetadata(
        report_id=report_id,
        entrypoint=_get_entrypoint(),
        arguments=_get_arguments(),
        start_time=start_time,
    )

    if options.save_reports:
        report_root = _get_report_root(
            override=str(options.report_directory) if options.report_directory else None
        )

        report_dir_name = f"{_REPORT_DIR_PREFIX}{report_id}"
        report_directory = report_root / report_dir_name
        while report_directory.exists():
            report_id = _generate_report_id(start_time)
            report_dir_name = f"{_REPORT_DIR_PREFIX}{report_id}"
            report_directory = report_root / report_dir_name
        report_directory.mkdir(parents=True, exist_ok=False)
        report_writer: ReportWriter = FileReportWriter(report_directory, report_summary)
    else:
        report_writer = NoOpReportWriter(report_summary)
    return report_writer


def filter_iteration_by_types(
    iteration: ReportIteration, filter_types: List[str]
) -> ReportIteration:
    """
    Filter a report iteration to only include objects whose types contain any of
    the filter strings.
    """
    if not filter_types:
        return iteration

    filtered_type_summaries = [
        ts
        for ts in iteration.leak_summary.type_summaries
        if any(filter_type.lower() in ts.object_type.lower() for filter_type in filter_types)
    ]

    filtered_referrers = [
        obj
        for obj in iteration.object_details_list
        if any(filter_type.lower() in obj.object_type_name.lower() for filter_type in filter_types)
    ]

    filtered_leak_summary = LeakSummary(
        iteration=iteration.leak_summary.iteration,
        type_summaries=filtered_type_summaries,
        max_types_in_summary=iteration.leak_summary.max_types_in_summary,
    )

    return ReportIteration(
        report_id=iteration.report_id,
        iteration_number=iteration.iteration_number,
        start_time=iteration.start_time,
        end_time=iteration.end_time,
        memory_usage=iteration.memory_usage,
        leak_summary=filtered_leak_summary,
        object_details_list=filtered_referrers,
    )


def get_memory_usage_output(
    previous: MemalotMemoryUsage | None, new: MemalotMemoryUsage
) -> MemoryUsageOutput:
    """
    Gets a `Output` object representing the memory usage. This includes the difference
    between the current and previous usage if the previous usage is provided.
    """
    return MemoryUsageOutput(
        current_rss_bytes=new.current_rss_bytes,
        peak_rss_bytes=new.peak_rss_bytes,
        current_rss_diff=new.current_rss_bytes - previous.current_rss_bytes if previous else None,
        system_percent_used=new.system_percent_used,
        peak_rss_diff=(
            new.peak_rss_bytes - previous.peak_rss_bytes
            if previous and previous.peak_rss_bytes is not None and new.peak_rss_bytes is not None
            else None
        ),
        diff_from_iteration=previous.iteration_number if previous else None,
    )


def get_report_reader(report_directory: Path | None = None) -> ReportReader:
    """
    Gets a report reader for the given report directory. If no directory is provided,
    the default report root is used.
    """
    return FileReportReader(report_directory=report_directory)


def _get_report_root(override: str | None) -> Path:
    """
    Gets the root directory for reports.
    """
    if override is None:
        report_root = _get_default_report_root()
    else:
        report_root = Path(override)
    return report_root


def _get_default_report_root() -> Path:
    """
    Gets the default report root.
    """
    return Path.home() / ".memalot" / "reports"


def _generate_report_id(start_time: datetime) -> str:
    """
    Generates a short, human-friendly, ID with a high probability of uniqueness
    (but the ID is not guaranteed to be unique, so needs to be checked).
    """
    chars = string.ascii_lowercase + string.digits
    first_part = "".join(random.choice(chars) for _ in range(4))
    second_part = "".join(random.choice(chars) for _ in range(4))
    return f"{first_part}-{second_part}"


def _get_entrypoint() -> str:
    """
    Gets the entrypoint of the report.
    """
    return sys.argv[0]


def _get_arguments() -> list[str]:
    """
    Gets the arguments of the report.
    """
    return sys.argv[1:]



import dataclasses
import gc
import inspect
import logging
from collections.abc import Callable
from typing import Any, Dict, Mapping, Type

import networkx as nx
from referrers import ReferrerGraphNode
from rich.markup import escape

from memalot.base import ReferrerGraph, ReferrerNode
from memalot.themes import (
    MEMORY_ABSOLUTE,
    OBJECT_ID,
    REFERRER_NAME,
    REFERRER_SUFFIX_CYCLE,
    REFERRER_SUFFIX_LEAF,
)

LOG = logging.getLogger(__name__)

_PACKAGE_PREFIX = "memalot"

_KIB = 1024
_MIB = _KIB**2
_GIB = _KIB**3


def as_mib(num_bytes: int, tag: str = MEMORY_ABSOLUTE) -> str:
    """
    Returns a string representation of the number of byte1s in MiB, rounded to one decimal
    place.
    """
    return f"[{tag}]{num_bytes / 1024 / 1024:.1f}[/{tag}] MiB"


def as_mib_sf(num_bytes: int, tag: str = MEMORY_ABSOLUTE) -> str:
    """
    Returns a string representation of the number of bytes in MiB, rounded to three significant
    figures.
    """
    return f"[{tag}]{num_bytes / 1024 / 1024:.3g}[/{tag}] MiB"


def format_bytes(num_bytes: int, tag: str = MEMORY_ABSOLUTE) -> str:
    """
    Converts a size in bytes to a human-readable string in B, KiB, MiB, or GiB.

    This function takes an integer representing the number of bytes and returns
    a formatted string with the most appropriate unit. The rounding is adjusted
    based on the unit.
    """
    if not isinstance(num_bytes, int) or num_bytes < 0:
        raise ValueError("Input must be a non-negative integer.")

    if num_bytes == 0:
        return "0 B"

    if num_bytes < _KIB:
        # For bytes, no decimal places are needed
        return f"[{tag}]{num_bytes}[/{tag}] B"
    elif num_bytes < _MIB:
        # For KiB, one decimal place is appropriate
        kib_value = num_bytes / _KIB
        return f"[{tag}]{kib_value:.1f}[/{tag}] KiB"
    elif num_bytes < _GIB:
        # For MiB, two decimal places offer good precision
        mib_value = num_bytes / _MIB
        return f"[{tag}]{mib_value:.2f}[/{tag}] MiB"
    else:
        # For GiB, two decimal places are also suitable
        gib_value = num_bytes / _GIB
        return f"[{tag}]{gib_value:.2f}[/{tag}] GiB"


def get_full_type_name(obj_type: Type[Any]) -> str:
    """
    Gets the full type name of an object.
    """
    return f"{obj_type.__module__}.{obj_type.__name__}"


def get_module_prefix() -> str | None:
    """
    Determine the module prefix from the top-level package of the calling code.

    This is the first package that is not part of Memalot.
    """
    stack_frames = inspect.stack()
    for frame_info in stack_frames:
        frame_module = inspect.getmodule(frame_info.frame)
        if frame_module and not frame_module.__name__.startswith(_PACKAGE_PREFIX):
            # Use the top-level package of the calling code as the module prefix
            # (with a trailing dot). For example, if the calling code is in a module
            # called my_module.do_thing, the module prefix would be "my_module.".
            # In some cases (like Jupyter notebooks), there may not be a top-level
            # package, in which there won't be any module prefixes. We log a warning
            # in this case.
            return f"{frame_module.__name__.split('.')[0]}."
    else:  # pragma: no cover
        LOG.warning(
            "Could not determine the top-level package of the calling code. "
            "You can specify the referrers_module_prefixes parameter to set this explicitly."
        )
        return None


def get_objects(
    max_untracked_search_depth: int = 3, get_objects_func: Callable[[], list[Any]] = gc.get_objects
) -> list[Any]:
    """
    Gets all objects that are currently in memory in the Python process, that are not eligible
    for garbage collection.

    This is different from `gc.get_objects` in a few ways:

     - It always performs a garbage collection when it is called.
     - It finds untracked objects, as long as they are referred to (directly or indirectly) by
       tracked objects. Untracked objects include, for example, mutable objects and collections
       containing only immutable objects in CPython. The `max_untracked_search_depth` controls
       how deep this function searches for untracked objects if they are not referred to directly
       by tracked objects. Setting this to a higher value is more likely to find untracked
       objects but will take more time.
     - It ignores frame objects.
     - It removes duplicate objects.

    :param max_untracked_search_depth: The maximum depth to search for untracked objects. This
        defaults to 3, which is enough to find most untracked objects. For example, this
        will find objects in tuples that are in another collection. However, it may not find
        certain untracked objects, like nested tuples.
    """
    gc.collect()
    tracked_objects = get_objects_func()

    all_objects = []
    seen_ids = set()

    for obj in tracked_objects:
        try:
            if not _is_excluded(obj):
                obj_id = id(obj)
                if obj_id not in seen_ids:
                    all_objects.append(obj)
                    seen_ids.add(obj_id)
        except ReferenceError:  # pragma: no cover
            # Some objects might not be accessible (e.g., weak reference proxies
            # where the underlying object has been garbage collected)
            pass

    # Search the referents of the objects we have found, looking for untracked objects.
    objects_to_search = list(all_objects)
    for _ in range(max_untracked_search_depth):
        new_untracked_referents = []
        for obj in objects_to_search:
            try:
                referents = gc.get_referents(obj)
                # Go through all referents and add them to new_untracked_referents if we
                # haven't seen them before, and they are untracked (and not excluded).
                for referent in referents:
                    referent_id = id(referent)
                    if (
                        referent_id not in seen_ids
                        and not gc.is_tracked(referent)
                        and not _is_excluded(referent)
                    ):
                        new_untracked_referents.append(referent)
                        all_objects.append(referent)
                        seen_ids.add(referent_id)
            except ReferenceError:  # pragma: no cover
                # Some objects might not be accessible
                pass
        objects_to_search = new_untracked_referents

    return all_objects


def _is_excluded(obj: Any) -> bool:
    try:
        return inspect.isframe(obj)
    except ReferenceError:  # pragma: no cover
        # This can happen if the object is a weak reference proxy where the underlying
        # object has been garbage collected. We just ignore these objects.
        return True


def convert_graph_nodes(
    referrer_graph: nx.DiGraph, replacements: Mapping[str, str] | None = None
) -> ReferrerGraph:
    """
    Converts a networkx graph with ReferrerGraphNode nodes to a ReferrerGraph
    with ReferrerNode nodes, transferring node attributes.
    """
    unique_ids: Dict[ReferrerGraphNode, int] = {}
    graph_nodes: list[ReferrerNode] = []
    target_ids: set[int] = set()
    root_ids: set[int] = set()

    # First pass: create nodes and assign unique IDs
    for unique_id, node in enumerate(referrer_graph.nodes()):
        unique_ids[node] = unique_id
        if isinstance(node, ReferrerGraphNode):
            name = node.name
            if replacements is not None:
                for old_name, new_name in replacements.items():
                    name = name.replace(old_name, new_name)
            # We ignore type here, as it doesn't seem to add anything that isn't present
            # in the name
            referrer_node = ReferrerNode(
                id=unique_id,
                name=name,
                object_id=node.id,
                referent_ids=[],  # Will be populated in second pass
            )
            graph_nodes.append(referrer_node)

            # Track target and root nodes separately
            if node.is_target:
                target_ids.add(unique_id)
            is_root = referrer_graph.in_degree(node) == 0
            if is_root:
                root_ids.add(unique_id)

        else:
            raise ValueError(f"Unexpected type: {type(node)}")

    # Second pass: populate referent_ids based on edges
    for u, v in referrer_graph.edges():
        if isinstance(u, ReferrerGraphNode) and isinstance(v, ReferrerGraphNode):
            u_id = unique_ids[u]
            v_id = unique_ids[v]
            graph_nodes[u_id].referent_ids.append(v_id)
        else:  # pragma: no cover
            raise ValueError(f"Unexpected type: {type(u)} or {type(v)}")

    return ReferrerGraph(target_ids=target_ids, root_ids=root_ids, graph_nodes=graph_nodes)


def pluralize(base_string: str, num: int) -> str:
    """
    Very simple pluralization. This just adds an "s" if `num` != 1.
    """
    if num == 1:
        return base_string
    else:
        return f"{base_string}s"


# Note: this is not a pydantic data class
@dataclasses.dataclass(frozen=True)
class PrintableReferrerNode:
    unique_id: int
    """
    A unique ID for the node.
    """

    name: str
    """
    A meaningful name for the referrer. For example, if the referrer is a local variable,
    the name would be the variable name, suffixed with "(local)".
    """

    object_id: int
    """
    A unique ID for the referrer object. If the referrer is not an object then this is the
    ID of the object it refers to.
    """

    is_cycle_member: bool = False
    """
    Whether the referrer is part of a cycle in the graph. If this is `True`, the referrer
    will be the last node in a branch of the graph.
    """

    is_root: bool = False
    """
    Whether this node is a root in the graph.
    """

    is_target: bool = False
    """
    Whether this node is the target node (i.e. the object we are finding referrers for).
    """

    def __str__(self) -> str:
        if self.is_cycle_member:
            suffix = f"[{REFERRER_SUFFIX_CYCLE}](cycle member)[/{REFERRER_SUFFIX_CYCLE}]"
        elif self.is_root:
            suffix = f"[{REFERRER_SUFFIX_LEAF}](root)[/{REFERRER_SUFFIX_LEAF}]"
        else:
            suffix = ""
        return (
            f"[{REFERRER_NAME}]{escape(self.name)}[/{REFERRER_NAME}] "
            f"[{OBJECT_ID}](id={self.object_id})[/{OBJECT_ID}] "
            f"{suffix}"
        )


def convert_graph_nodes_to_printable(referrer_graph: ReferrerGraph) -> nx.DiGraph:
    """
    Converts a ReferrerGraph with ReferrerNode nodes, as produced by
    `utils.convert_graph_nodes` to a graph with `_PrintableReferrerNode`s.

    Cycles in the graph are broken by nodes with multiple children and marking all duplicated
    nodes as cycle members (except for the first, original, instance of the node).

    The edges in the returned graph are reversed, so that they point from the target object
    to the root objects (from referents to referrers). This makes it easier to visualise the
    graph when printed.
    """
    id_sequence = max(node.id for node in referrer_graph.graph_nodes) + 1

    id_graph = nx.DiGraph()

    # First pass: add all nodes to the ID graph to the new graph
    for node in referrer_graph.graph_nodes:
        id_graph.add_node(node.id, data=node, original_node_id=node.id)

    # Second pass: add edges to the ID graph, breaking cycles
    for node in referrer_graph.graph_nodes:
        # If this node has more than one referent, create a copy of the node for all but the
        # first referent, so that the graph is a tree. Mark the copies as cycle members.
        if node.referent_ids:
            id_graph.add_edge(node.id, node.referent_ids[0])
            for referrer_id in node.referent_ids[1:]:
                new_node = dataclasses.replace(node, id=id_sequence)
                id_graph.add_node(
                    id_sequence, data=new_node, is_cycle_member=True, original_node_id=node.id
                )
                id_graph.add_edge(id_sequence, referrer_id)
                id_sequence += 1

    # Now, create the printable graph, reversing the edges
    printable_graph = nx.DiGraph()
    id_to_printable_node: Dict[int, PrintableReferrerNode] = {}
    for node_id in id_graph.nodes():
        node_data = id_graph.nodes[node_id]["data"]
        is_cycle_member = id_graph.nodes[node_id].get("is_cycle_member", False)
        original_node_id = id_graph.nodes[node_id].get("original_node_id")
        is_root = original_node_id in referrer_graph.root_ids
        is_target = original_node_id in referrer_graph.target_ids
        printable_node = PrintableReferrerNode(
            unique_id=node_data.id,
            name=node_data.name,
            object_id=node_data.object_id,
            is_cycle_member=is_cycle_member,
            is_root=is_root,
            is_target=is_target,
        )
        printable_graph.add_node(printable_node)
        id_to_printable_node[node_id] = printable_node
    for u, v in id_graph.edges():
        printable_graph.add_edge(id_to_printable_node[u], id_to_printable_node[v])
    printable_graph = printable_graph.reverse(copy=False)

    return printable_graph


def get_object_type_budgets(
    object_type_counts: Mapping[type, int], total_budget: int
) -> Mapping[type, int]:
    """
    Calculate the budget for the most common object types.

    Each object type will get a budget of at least one. The remaining budget
    (based on `total_budget`) is distributed proportionally to the
    number of occurrences of each type.

    For example, if `total_budget` is set to 10 and type1 occurs 20 times and type2 occurs 4
    times, then the budget for type1 will be 8 and the budget
    for type2 will be 2.

    If there are as many types as `max_object_details`, then the budget for each
    type will be one.
    """
    num_types = len(object_type_counts)
    if num_types == 0:
        return {}

    if num_types >= total_budget:
        return dict.fromkeys(object_type_counts, 1)

    budget = dict.fromkeys(object_type_counts, 1)
    remaining_budget = total_budget - num_types

    total_count = sum(object_type_counts.values())

    for obj_type, count in object_type_counts.items():
        proportional_budget = round((count / total_count) * remaining_budget)
        budget[obj_type] += proportional_budget

    return budget



from dataclasses import field
from pathlib import Path
from typing import AbstractSet, Any, Callable, Optional

from pydantic.dataclasses import dataclass


class InvalidOptionsError(Exception):
    """
    Raised when an option is invalid.
    """


@dataclass(frozen=True)
class Options:
    """
    Options for controlling the behavior of memalot. These are passed to the API as keyword
    arguments. For example::

        memalot.start(max_object_lifetime=60.0, force_terminal=True)

    In this case, the `force_terminal` argument matches a property in this class.
    """

    included_type_names: AbstractSet[str] = field(default_factory=set)
    """
    The types of objects to include in the report. By default all types are checked,
    but this can be limited to a subset of types.

    Inclusion is based on substring matching of the fully-qualified type name (the name of
    the type and its module). For example, if `included_type_names` is set to `{"numpy"}`,
    *all* NumPy types will be included in the report.
    """

    excluded_type_names: AbstractSet[str] = field(default_factory=set)
    """
    The types of objects to exclude from the report. By default no types are excluded.

    Exclusion is based on substring matching of the fully-qualified type name (the name of
    the type and its module). For example, if `excluded_type_names` is set to `{"numpy"}`,
    *all* NumPy types will be excluded from the report.
    """

    max_types_in_leak_summary: int = 500
    """
    The maximum number of types to include in the leak summary.
    """

    compute_size_in_leak_summary: bool = False
    """
    Computes the (shallow) size of all objects in the leak summary.

    Note: the shallow size of an object may not be particularly meaningful, since most
    objects refer to other objects, and often don't contain much data themselves.
    """

    max_untracked_search_depth: int = 3
    """
    The maximum search depth when looking for leaked objects that are not tracked by the
    garbage collector. Untracked objects include, for example, mutable objects and collections
    containing only immutable objects in CPython

    This defaults to 3, which is enough to find most untracked objects. However, this may not
    be sufficient to find some untracked objects, like nested tuples. Increase this if you
    have nested collections of immutable objects (like tuples). However, note that increasing
    this may impact speed.
    """

    check_referrers: bool = True
    """
    Whether to check for referrers of leaked objects. Defaults to `True`.

    This option may cause a significant slow-down (but provides useful information).
    Try setting this to `False` if Memalot is taking a long time to generate object details.
    Then, when you have an idea of what types of objects are leaking, you can generate reports
    with `check_referrers=True` and `included_type_names` set to the types of objects that you
    think may be leaking.
    """

    max_object_details: int = 30
    """
    The maximum number of objects for which to print details. Defaults to `30`.

    We try to check at least one object for each object type, within this limit. If the
    number of types exceeds this limit, then we check only the most common types. If the
    number of types is less than this limit, then we will check more objects for more
    common types.
    """

    referrers_max_depth: int | None = 50
    """
    The maximum depth to search for referrers. The default is 50. Specify `None` to search to
    unlimited depth (but be careful with this: it may take a long time).
    """

    referrers_search_timeout: float | None = 60.0 * 5
    """
    The maximum time to spend searching for referrers for an individual object. If this time
    is exceeded, a partial graph is displayed and the referrer graph will contain a node
    containing the text "Timeout of N seconds exceeded". Note that this timeout is approximate,
    and may not be effective if the search is blocked by a long-running operation. The default
    is 5 minutes. Setting this to `None` will disable the timeout.
    """

    single_object_referrer_limit: int | None = 100
    """
    The maximum number of referrers to include in the graph for an individual object instance.
    If the limit is exceeded, the referrer graph will contain a node containing the text
    "Referrer limit of N exceeded". Note that this limit is approximate and does not apply to all
    referrer types. Specifically, it only applies to object references. Additionally,
    this limit does not apply to immortal objects.
    """

    referrers_module_prefixes: Optional[AbstractSet[str]] = None
    """
    The prefixes of the modules to search for module-level variables when looking for referrers.
    If this is not specified, the top-level package of the calling code is used.
    """

    referrers_max_untracked_search_depth: int = 30
    """
    The maximum depth to search for referrers of untracked objects. This is the depth that
    referents will be searched from the roots (locals and globals). The default is 30. If you
    are missing referrers of untracked objects, you can increase this value.
    """

    save_reports: bool = True
    """
    Whether to save reports to disk. This is useful for inspecting them later.

    Reports are written to the `report_directory`, or the default directory if
    this is not specified.
    """

    report_directory: Path | None = None
    """
    The directory to write the report data to. Individual report data is written to
    a subdirectory of this directory.

    If this is `None` (the default), the default directory will be used. This is the
    `.memalot/reports` directory in the user's home directory.

    To turn off saving of reports entirely, use the `save_reports` option.
    """

    str_func: Callable[[Any, int], str] | None = None
    """
    A function for outputting the string representation of an object. The first argument
    is the object and the second argument is the length to truncate the string to,
    as specified by `str_max_length`.

    If this is not supplied the object's `__str__` is used.
    """

    str_max_length: int = 100
    """
    The maximum length of object string representations, as passed to `str_func`.
    """

    force_terminal: bool | None = None
    """
    Forces the use of terminal control codes, which enable colors and other formatting.
    Defaults to `False`, as this is normally detected automatically.

    Set this to `True` if you are missing colors or other formatting in the output, as
    sometimes (like when running in an IDE) the terminal is not detected correctly.

    This must be set to `False` if `output_func` is set and `tee_console` is `False`.
    """

    output_func: Callable[[str], None] | None = None
    """
    A function that writes reports. If this is not provided reports are
    printed to the console.

    This option can be used to, for example, write reports to a log file.

    If this option is specified then output is *not* written to the console, unless
    `tee_console` is set to `True`.
    """

    tee_console: bool = False
    """
    If this is set to `True`, output is written to the console as well as to the function
    specified by `output_func`. If `output_func` is not specified (the default) then this
    option has no effect.
    """

    color: bool = True
    """
    Specifies whether colors should be printed to the console.

    Note: in certain consoles (like when running in an IDE), colors are not printed by
    default. Try setting `force_terminal` to `True` if this happens.
    """

    def __post_init__(self) -> None:
        if self.output_func is not None and not self.tee_console and self.force_terminal:
            raise InvalidOptionsError(
                "If output_func is specified and tee_console is False, then force_terminal "
                "must be False."
            )



import collections
from abc import ABC, abstractmethod
from typing import Any, Generic, Iterable, Iterator, List, TypeVar

from pydantic.dataclasses import dataclass

_HASHABLE_TYPE = collections.abc.Hashable


class MemalotInt(int):
    """
    An int. This is used in preference to int so that it can be excluded from reports.
    """

    pass


class MemalotCount(int):
    """
    A count. This is used in preference to int so that it can be excluded from reports.
    """

    pass


class MemalotObjectId(int):
    """
    An object ID. This is used in preference to int so that it can be excluded from reports.
    """


class MemalotObjectIds(set[int]):
    """
    A set of object IDs. This is used in preference to a set so that it can be excluded
    from reports.
    """


class MemalotList(list[Any]):
    """
    A list. This is used in preference to a regular list so that it can be excluded from reports.
    """


class MemalotSet(set[Any]):
    """
    A set. This is used in preference to a regular list so that it can be excluded from reports.
    """


class ObjectSignature:
    """
    An object signature that uses heuristics to determine whether two objects are the same
    or not.

    Objects cannot be compared by their ID alone, since after objects are garbage collected
    their IDs may be reused (from the Python docs: "objects with non-overlapping lifetimes may
    have the same id() value") so this class uses the object's type and hash as well as the
    object ID. If objects are not hashable, then only the object ID and type are used.
    """

    def __init__(self, obj: Any) -> None:
        self._id = MemalotObjectId(id(obj))
        self._type = type(obj)
        self._signature = self._get_hash(obj)

    def _get_hash(self, obj: Any) -> MemalotInt:
        try:
            return self._hash(obj)
        except Exception:
            pass
        # Fall-back to using the object ID if the object is not hashable (or another
        # error occurs getting the hash). This isn't great, since it won't work for collections
        # like dicts and lists, but there's probably not much we can do about that
        # (they're mutable anyway, so comparing their contents won't help us much).
        return MemalotInt(id(self))

    def _hash(self, obj: Any) -> MemalotInt:
        return MemalotInt(hash(obj))

    def is_probably_same_object(self, obj: Any) -> bool:
        return (
            MemalotObjectId(id(obj)) == self._id and type(obj) is self._type and self._get_hash(obj)
        ) == self._signature


@dataclass
class ApproximateSize:
    """
    Represents an approximate uncertain size. That is, a size where the lower bound is known
    approximately, and the upper bound may not be known at all.
    """

    approx_size: int = 0
    """
    If `upper_bound_known` is `True`, then this is the approximate size. If `upper_bound_known` is
    `False`, then this is the approximate lower bound of the size.
    """

    upper_bound_known: bool = True
    """
    Whether the upper bound of the size is known.
    """

    def __add__(self, other: "ApproximateSize") -> "ApproximateSize":
        return ApproximateSize(
            self.approx_size + other.approx_size, self.upper_bound_known and other.upper_bound_known
        )

    @property
    def prefix(self) -> str:
        if self.upper_bound_known:
            return "~"
        else:
            return ">="

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ApproximateSize):
            return False
        return (
            self.approx_size == other.approx_size
            and self.upper_bound_known == other.upper_bound_known
        )

    def __hash__(self) -> int:
        return hash((self.approx_size, self.upper_bound_known))


T = TypeVar("T")


class CachingIterable(Generic[T]):
    """
    Iterable that caches the underlying iterable.

    Note: this class is *not* thread-safe!
    """

    def __init__(self, iterable: Iterable[T]):
        self._iterator = iter(iterable)
        self._cache: List[T] = []

    def __iter__(self) -> Iterator[T]:
        for item in self._cache:
            yield item
        try:
            while True:
                item = next(self._iterator)
                self._cache.append(item)
                yield item
        except StopIteration:
            pass

    def __len__(self) -> int:
        # Consume the entire iterator to get the length
        # We need to exhaust the iterator without calling list(self) to avoid recursion
        for _ in self:
            pass
        return len(self._cache)


@dataclass
class ReferrerNode:
    """
    Represents a node in a referrer graph.

    This represents something that refers to other objects. For example, a local variable,
    an instance attribute, a collection, or a global variable. In addition, objects themselves
    can be represented in the graph as nodes.
    """

    id: int
    """
    A unique ID for the node.
    """

    name: str
    """
    A descriptive (human-readable) name for the node.
    """

    object_id: int
    """
    The object ID that the node represents. Note that this is not a unique ID, since multiple
    nodes can represent the same object. For example, an object may be represented by
    (potentially multiple) instance attributes, and by the object itself.
    """

    referent_ids: list[int]
    """
    The IDs of the referents of the node. These are the nodes that this node refers to.

    For example, if the node represents a local variable, then the referent is the object
    that the variable refers to.
    """


@dataclass
class ReferrerGraph:
    """
    A graph of referrers.

    This graph is used to represent the referrers (both direct and indirect) of one or more
    target objects.
    """

    target_ids: set[int]
    """
    IDs of the target objects. These are the objects that we have found the referrers of.
    """

    root_ids: set[int]
    """
    IDs of the root nodes. These are nodes that have no referrers.
    """

    graph_nodes: list[ReferrerNode]
    """
    The nodes in the graph.

    Edges are represented by the `referent_ids` attribute of the nodes.
    """

    def __len__(self) -> int:
        return len(self.graph_nodes)


class ObjectGetter(ABC):
    """
    Gets objects in memory.
    """

    @abstractmethod
    def get_objects(self) -> list[Any]:
        """
        Get filtered objects by calling the underlying function and applying filters.
        """
        pass  # pragma: no cover
